#!/usr/bin/env python
# -*- coding: utf-8 -*-

import decimal
import unittest
import datetime

from sane_finances.sources.base import ParseError
from sane_finances.sources.msci.v1.meta import IndexValue, IndexInfo, Styles, Sizes
from sane_finances.sources.msci.v1.parsers import MsciHistoryXmlParser, MsciIndexInfoParser


class TestMsciHistoryXmlParser(unittest.TestCase):

    def test_parse_Success(self):
        parser = MsciHistoryXmlParser()

        expected_result = [IndexValue(
            date=datetime.date(2016, 11, 3),
            value=decimal.Decimal('1644.941'),
            index_name='EAFE',
            style=Styles.NONE,
            size=Sizes.REGIONAL_STANDARD)]

        expected_name, expected_style, expected_size = \
            expected_result[0].index_name, expected_result[0].style.value, expected_result[0].size.value
        expected_date_str = expected_result[0].date.strftime(parser.date_format)
        expected_value_str = f"{expected_result[0].value:,}"  # comma as separator for thousands

        valid_xml = f"""<?xml version="1.0" ?>  <performance>
        <index id="{expected_name},{expected_style},{expected_size}">
          <asOf>
            <date>{expected_date_str}</date>
            <value>{expected_value_str}</value>
          </asOf>
        </index>
      </performance>"""

        result = list(parser.parse(valid_xml, tzinfo=None))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_raisesWhenNoData(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = ''

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenEmptyList(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance></performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNotList(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<WRONG_TAG></WRONG_TAG>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenWrongIndexId(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance>
          <index id="WRONG">
            <asOf>
              <date>03/11/2016</date>
              <value>1,644.941</value>
            </asOf>
          </index>
        </performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoDateTag(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance>
          <index id="EAFE,C,36">
            <asOf>
              <date__>03/11/2016</date__>
              <value>1,644.941</value>
            </asOf>
          </index>
        </performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoValueTag(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance>
          <index id="EAFE,C,36">
            <asOf>
              <date>03/11/2016</date>
              <value__>1,644.941</value__>
            </asOf>
          </index>
        </performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenWrongDate(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance>
          <index id="EAFE,C,36">
            <asOf>
              <date>9999/9999/9999</date>
              <value>1,644.941</value>
            </asOf>
          </index>
        </performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenWrongValue(self):
        parser = MsciHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" ?>  <performance>
          <index id="EAFE,C,36">
            <asOf>
              <date>03/11/2016</date>
              <value>1.644.941</value>
            </asOf>
          </index>
        </performance>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))


class TestMsciIndexInfoParser(unittest.TestCase):

    def test_parse_success(self):
        expected_result = [IndexInfo(index_id='2670', name='AC AMERICAS')]

        expected_name, expected_id = expected_result[0].name, expected_result[0].index_id

        xml = f"""<?xml version="1.0" ?><indices><index id="{expected_id}" name="{expected_name}" /></indices>
        """

        parser = MsciIndexInfoParser()

        result = list(parser.parse(xml))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_raisesEmptyList(self):
        xml = """<?xml version="1.0" ?><indices></indices>"""

        parser = MsciIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutName(self):
        xml = """<?xml version="1.0" ?><indices><index id="2670" /></indices>"""

        parser = MsciIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutId(self):
        xml = """<?xml version="1.0" ?><indices><index name="AC AMERICAS" /></indices>"""

        parser = MsciIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWrongTagName(self):
        xml = """<?xml version="1.0" ?><WRONG_TAG><index id="2670" name="AC AMERICAS" /></WRONG_TAG>"""

        parser = MsciIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesEmptyString(self):
        xml = ''

        parser = MsciIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tools for download (export) index history data from solactive.com
"""

import datetime
import decimal
import logging
import urllib.parse
import typing

from .meta import (
    IndexHistoryTypes, SolactiveIndexHistoryDownloadParameters,
    SolactiveIndexesInfoDownloadParameters, SolactiveDownloadParametersFactory)
from .parsers import SolactiveJsonParser, SolactiveIndexInfoParser
from ....communication.downloader import Downloader, DownloadStringResult
from ...base import (
    ApiActualityChecker, InstrumentStringDataDownloader, InstrumentExporterFactory,
    InstrumentHistoryValuesExporter, InstrumentsInfoExporter,
    DownloadParameterValuesStorage, ParseError,
    CheckApiActualityError)
from ...generic import GenericInstrumentHistoryValuesExporter, GenericInstrumentsInfoExporter

logging.getLogger().addHandler(logging.NullHandler())


class SolactiveStringDataDownloader(InstrumentStringDataDownloader):
    """ Data downloader from Solactive.com.
    """

    IndexHistoryUrl = 'https://www.solactive.com/indices/'

    def __init__(self, downloader: Downloader):
        """
        """
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader

        # headers HTTP
        self.headers = {}

        # query parameters
        self.index_history_type = IndexHistoryTypes.MAX

    def download_instrument_history_string(
            self,
            parameters: SolactiveIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        return self.download_index_history_string(parameters.isin)

    def download_instruments_info_string(
            self,
            parameters: SolactiveIndexesInfoDownloadParameters) -> DownloadStringResult:
        return self.download_all_indexes_info_string()

    def download_index_history_string(self, isin: str) -> DownloadStringResult:
        """ Downloads data for one index as string.

        :param isin: ISIN.
        :return: Container with downloaded string.
        """
        params = [
            ('indexhistory', isin),  # index ID
            ('indexhistorytype', self.index_history_type.value),  # period
        ]

        headers = dict(self.headers)  # make a copy

        # additional headers, mandatory for REST API operability
        headers.update({'Referer': self.IndexHistoryUrl + '?' + urllib.parse.urlencode({'index': isin})})

        self.downloader.parameters = params
        self.downloader.headers = headers

        return self.downloader.download_string(self.IndexHistoryUrl)

    def download_all_indexes_info_string(self) -> DownloadStringResult:
        """ Downloads string with the list of all available indexes

        :return: Container with downloaded string.
        """
        self.downloader.parameters = []
        self.downloader.headers = {}

        return self.downloader.download_string(self.IndexHistoryUrl)


class SolactiveApiActualityChecker(ApiActualityChecker):
    """ Verifies actuality and accessibility of REST API of solactive.com.
    """

    ActualityCheckISIN = 'DE000SLA4YD9'  # FXUS, Solactive GBS United States Large & Mid Cap Index NTR

    _expectedFirstMomentDate = datetime.date(2006, 5, 8)
    _expectedFirstValue = decimal.Decimal('466.44')

    def __init__(self,
                 string_data_downloader: SolactiveStringDataDownloader,
                 json_parser: SolactiveJsonParser):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.string_data_downloader = string_data_downloader
        self.json_parser = json_parser

    def check(self):
        expected_isin = self.ActualityCheckISIN

        self.logger.info(f"Check actuality via {expected_isin!r}")

        str_data_result = self.string_data_downloader.download_index_history_string(expected_isin)
        self.logger.debug(f"Got str data: {str_data_result.downloaded_string}")

        try:
            data = list(self.json_parser.parse(str_data_result.downloaded_string, tzinfo=None))
        except ParseError as ex:
            str_data_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected index history JSON: {ex.message}") from ex
        except Exception:
            str_data_result.set_correctness(False)
            raise

        data.sort(key=lambda v: v.moment)

        first_value = data[0]  # for actuality checking one record is enough

        if first_value.index_id != expected_isin:
            str_data_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected index ID. {first_value.index_id!r} != {expected_isin!r}")

        if first_value.moment.date() != self._expectedFirstMomentDate:
            str_data_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first date. {first_value.moment.date()} != {self._expectedFirstMomentDate}")

        if first_value.value != self._expectedFirstValue:
            str_data_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first value. {first_value.value} != {self._expectedFirstValue}")

        self.logger.info("Actuality check was successful")


class SolactiveDownloadParameterValuesStorage(DownloadParameterValuesStorage):
    """ Storage of indexes download parameters.
    """

    def is_dynamic_enum_type(self, cls: type) -> bool:
        return False

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return ()

    def get_dynamic_enum_key(self, instance):
        return None


class SolactiveExporterFactory(InstrumentExporterFactory):
    """ Factory class for create instances of Solactive data exporter.
    """
    name: str = 'Solactive index data exporter. Version 2018.'
    provider_site: str = 'https://www.solactive.com/'

    def __init__(self):
        self._dynamic_enum_type_manager = None
        self._download_parameters_factory = None

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        string_data_downloader = SolactiveStringDataDownloader(downloader)
        history_values_parser = SolactiveJsonParser()

        return GenericInstrumentHistoryValuesExporter(string_data_downloader, history_values_parser)

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        string_data_downloader = SolactiveStringDataDownloader(downloader)
        info_parser = SolactiveIndexInfoParser()

        return GenericInstrumentsInfoExporter(string_data_downloader, info_parser)

    def create_download_parameter_values_storage(self, downloader: Downloader) -> DownloadParameterValuesStorage:
        return SolactiveDownloadParameterValuesStorage()

    def create_api_actuality_checker(self, downloader: Downloader) -> SolactiveApiActualityChecker:
        string_data_downloader = SolactiveStringDataDownloader(downloader)
        history_values_parser = SolactiveJsonParser()

        return SolactiveApiActualityChecker(string_data_downloader, history_values_parser)

    @property
    def dynamic_enum_type_manager(self) -> SolactiveDownloadParameterValuesStorage:
        if self._dynamic_enum_type_manager is None:
            self._dynamic_enum_type_manager = SolactiveDownloadParameterValuesStorage()

        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> SolactiveDownloadParametersFactory:
        if self._download_parameters_factory is None:
            self._download_parameters_factory = SolactiveDownloadParametersFactory()

        return self._download_parameters_factory



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import decimal
import unittest
import datetime

from sane_finances.sources.base import ParseError
from sane_finances.sources.cbr.v2016.meta import CurrencyInfo, CurrencyRateValue
from sane_finances.sources.cbr.v2016.parsers import CbrCurrencyHistoryXmlParser, CbrCurrencyInfoParser


class TestCbrCurrencyHistoryXmlParser(unittest.TestCase):

    def test_parse_Success(self):
        parser = CbrCurrencyHistoryXmlParser()

        expected_result = [CurrencyRateValue(
            date=datetime.date(2000, 1, 1),
            value=decimal.Decimal("27.0100"),
            nominal=1,
            currency_id="R01235")]

        expected_id, expected_nominal_str = expected_result[0].currency_id, str(expected_result[0].nominal)
        expected_date_str = expected_result[0].date.strftime(parser.date_format)
        expected_value_str = f"{expected_result[0].value}".replace('.', ',')

        valid_xml = f"""<?xml version="1.0" encoding="windows-1251"?>
<ValCurs ID="{expected_id}" DateRange1="01.01.2000" DateRange2="01.02.2000" name="Foreign Currency Market Dynamic">
    <Record Date="{expected_date_str}" Id="{expected_id}">
        <Nominal>{expected_nominal_str}</Nominal>
        <Value>{expected_value_str}</Value>
        </Record>
</ValCurs>"""

        result = list(parser.parse(valid_xml, tzinfo=None))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_raisesWhenNoData(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = ''

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_WhenEmptyList(self):
        parser = CbrCurrencyHistoryXmlParser()

        valid_xml = f"""<?xml version="1.0" encoding="windows-1251"?>
<ValCurs ID="R01235" DateRange1="02.01.2000" DateRange2="02.01.2000" name="Foreign Currency Market Dynamic"/>
"""
        result = list(parser.parse(valid_xml, tzinfo=None))

        self.assertEqual(len(result), 0)

    def test_parse_raisesWhenNotList(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" encoding="windows-1251"?><WRONG_TAG></WRONG_TAG>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoDateAttrib(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" encoding="windows-1251"?>
        <ValCurs ID="R01235" DateRange1="01.01.2000" DateRange2="01.02.2000" name="Foreign Currency Market Dynamic">
            <Record Date__="01.01.2000" Id="R01235">
                <Nominal>1</Nominal>
                <Value>27,0000</Value>
            </Record>
        </ValCurs>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoIdAttrib(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" encoding="windows-1251"?>
        <ValCurs ID="R01235" DateRange1="01.01.2000" DateRange2="01.02.2000" name="Foreign Currency Market Dynamic">
            <Record Date="01.01.2000" Id__="R01235">
                <Nominal>1</Nominal>
                <Value>27,0000</Value>
            </Record>
        </ValCurs>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoNominalTag(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" encoding="windows-1251"?>
        <ValCurs ID="R01235" DateRange1="01.01.2000" DateRange2="01.02.2000" name="Foreign Currency Market Dynamic">
            <Record Date="01.01.2000" Id="R01235">
                <Nominal__>1</Nominal__>
                <Value>27,0000</Value>
            </Record>
        </ValCurs>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))

    def test_parse_raisesWhenNoValueTag(self):
        parser = CbrCurrencyHistoryXmlParser()

        invalid_xml = """<?xml version="1.0" encoding="windows-1251"?>
        <ValCurs ID="R01235" DateRange1="01.01.2000" DateRange2="01.02.2000" name="Foreign Currency Market Dynamic">
            <Record Date="01.01.2000" Id="R01235">
                <Nominal>1</Nominal>
                <Value__>27,0000</Value__>
            </Record>
        </ValCurs>"""

        with self.assertRaises(ParseError):
            list(parser.parse(invalid_xml, tzinfo=None))


class TestCbrCurrencyInfoParser(unittest.TestCase):

    def test_parse_success(self):
        expected_result = [CurrencyInfo(
            currency_id='R01010',
            name='Австралийский доллар',
            eng_name='Australian Dollar',
            nominal=1,
            parent_code='R01010')]

        expected_name, expected_id = expected_result[0].name, expected_result[0].currency_id
        expected_eng_name = expected_result[0].eng_name
        expected_nominal = expected_result[0].nominal
        expected_parent_code = expected_result[0].parent_code

        xml = f"""<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
    <Item ID="{expected_id}">
        <Name>{expected_name}</Name>
        <EngName>{expected_eng_name}</EngName>
        <Nominal>{expected_nominal}</Nominal>
        <ParentCode>{expected_parent_code} </ParentCode>
    </Item>
</Valuta>
        """

        parser = CbrCurrencyInfoParser()

        result = list(parser.parse(xml))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_raisesWithoutName(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
            <Item ID="R01010">
                <Name__>Австралийский доллар</Name__>
                <EngName>Australian Dollar</EngName>
                <Nominal>1</Nominal>
                <ParentCode>R01010 </ParentCode>
            </Item>
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutEngName(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
            <Item ID="R01010">
                <Name>Австралийский доллар</Name>
                <EngName__>Australian Dollar</EngName__>
                <Nominal>1</Nominal>
                <ParentCode>R01010 </ParentCode>
            </Item>
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutNominal(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
            <Item ID="R01010">
                <Name>Австралийский доллар</Name>
                <EngName>Australian Dollar</EngName>
                <Nominal__>1</Nominal__>
                <ParentCode>R01010 </ParentCode>
            </Item>
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutParentCode(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
            <Item ID="R01010">
                <Name>Австралийский доллар</Name>
                <EngName>Australian Dollar</EngName>
                <Nominal>1</Nominal>
                <ParentCode__>R01010 </ParentCode__>
            </Item>
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWithoutIdTag(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
            <Item ID__="R01010">
                <Name>Австралийский доллар</Name>
                <EngName>Australian Dollar</EngName>
                <Nominal>1</Nominal>
                <ParentCode__>R01010 </ParentCode__>
            </Item>
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesWrongTagName(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><WRONG_TAG name="Foreign Currency Market Lib">
    <Item ID="R01010">
        <Name>Австралийский доллар</Name>
        <EngName>Australian Dollar</EngName>
        <Nominal>1</Nominal>
        <ParentCode>R01010 </ParentCode>
    </Item>
</WRONG_TAG>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesEmptyList(self):
        xml = """<?xml version="1.0" encoding="windows-1251"?><Valuta name="Foreign Currency Market Lib">
        </Valuta>"""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))

    def test_parse_raisesEmptyString(self):
        xml = ""

        parser = CbrCurrencyInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(xml))



#!/usr/bin/python
# -*- coding: utf-8 -*-

""" Annotations for instance attributes, functions arguments, etc.
"""

import sys
import operator
import collections.abc
import logging
import typing

logging.getLogger().addHandler(logging.NullHandler())

T = typing.TypeVar('T')  # pylint: disable=invalid-name


@typing.runtime_checkable
class SupportsDescription(typing.Protocol):
    """ An ABC with description attribute of type string.

    Can be used with ``isinstance`` to check whether object contains such attribute.
    Supposed for using with custom enums in meta modules.
    """
    description: str


class Description:
    """ Description for attribute's or function argument's annotation

    ``description`` can be used for wide description.

    ``short_description`` can be used for short description, e.g. for label of input in HTML form.
    """
    description: str
    short_description: str = None

    def __init__(self, description: str, short_description: str = None):
        """ Initialize description.

        :param description: Description of attribute or argument.
        :param short_description: Short description (optional).
        """
        self.description = str(description)
        self.short_description = None if short_description is None else str(short_description)


class Volatile:
    """ Describes attribute or function argument which value even through not optional,
    but can be reevaluated each time when uses for instantiation or function calling.

    For example: some source demands interval [date_from; date_to] for download history data,
    but we need to download it every day. So we store this interval somewhere in configuration
    and mark date_to as ``Volatile`` with reevaluating as ``date.today()``
    every time we fetch parameters from configuration.
    Thus interval widened every day without modifying configuration.

    ::

        class Params:
            isin: str
            date_from: datetime.date
            date_to: Annotated[datetime.date, Volatile(lambda ctx: datetime.date.today(), datetime.date.max)]
    """

    def __init__(self,
                 generator: typing.Callable[[typing.Dict[str, typing.Any]], T],
                 stub_value: T = None):
        """ Initialize volatile annotation.

        :param generator: Function for reevaluating;
                          takes context of instantiation or function calling (like ``kwargs``)
                          and returns new value of attribute.
        :param stub_value: Stub value for not optional attributes.
        """
        if not callable(generator):
            raise TypeError("generator is not callable")

        self.generator = generator
        self.stub_value = stub_value

    def generate(self, context: typing.Dict[str, typing.Any]) -> T:
        """ Generate actual value of attribute or function argument based on context.

        :param context: Context of instantiation or function calling (like ``kwargs``)
        :return: Actual value of volatile attribute.
        """
        return self.generator(context)


# Support legacy 3.8 version:
# Got from https://github.com/python/typing (typing_extensions module)

# After PEP 560, internal typing API was substantially reworked.
# This is especially important for Protocol class which uses internal APIs
# quite extensivelly.
PEP_560 = sys.version_info[:3] >= (3, 7, 0)

try:  # pragma: no cover
    # noinspection PyProtectedMember
    from typing import _tp_cache
except ImportError:  # pragma: no cover
    def _tp_cache(x):  # pylint: disable=invalid-name
        return x

LEGACY_ANNOTATIONS = True  # flag about legacy python version

# Python 3.9+ has PEP 593 (Annotated and modified get_type_hints)
if hasattr(typing, 'Annotated'):  # pragma: no cover

    Annotated = typing.Annotated
    get_type_hints = typing.get_type_hints

    # Not exported and not a public API, but needed for get_origin() and get_args()
    # to work.
    # noinspection PyUnresolvedReferences,PyProtectedMember
    _AnnotatedAlias = typing._AnnotatedAlias

    LEGACY_ANNOTATIONS = False

elif PEP_560:  # pragma: no cover

    # noinspection PyUnresolvedReferences,PyProtectedMember
    class _AnnotatedAlias(typing._GenericAlias, _root=True):
        """Runtime representation of an annotated type.
        At its core 'Annotated[t, dec1, dec2, ...]' is an alias for the type 't'
        with extra annotations. The alias behaves like a normal typing alias,
        instantiating is the same as instantiating the underlying type, binding
        it to types is also the same.
        """
        def __init__(self, origin, metadata):
            if isinstance(origin, _AnnotatedAlias):
                metadata = origin.__metadata__ + metadata
                origin = origin.__origin__
            super().__init__(origin, origin)
            self.__metadata__ = metadata

        def copy_with(self, params):
            assert len(params) == 1
            new_type = params[0]
            return _AnnotatedAlias(new_type, self.__metadata__)

        def __repr__(self):
            return "typing_extensions.Annotated[{}, {}]".format(
                typing._type_repr(self.__origin__),
                ", ".join(repr(a) for a in self.__metadata__)
            )

        def __reduce__(self):
            return operator.getitem, (
                Annotated, (self.__origin__,) + self.__metadata__
            )

        def __eq__(self, other):
            if not isinstance(other, _AnnotatedAlias):
                return NotImplemented
            if self.__origin__ != other.__origin__:
                return False
            return self.__metadata__ == other.__metadata__

        def __hash__(self):
            return hash((self.__origin__, self.__metadata__))

    class Annotated:
        """Add context specific metadata to a type.
        Example: Annotated[int, runtime_check.Unsigned] indicates to the
        hypothetical runtime_check module that this type is an unsigned int.
        Every other consumer of this type can ignore this metadata and treat
        this type as int.
        The first argument to Annotated must be a valid type (and will be in
        the __origin__ field), the remaining arguments are kept as a tuple in
        the __extra__ field.
        Details:
        - It's an error to call `Annotated` with less than two arguments.
        - Nested Annotated are flattened::
            Annotated[Annotated[T, Ann1, Ann2], Ann3] == Annotated[T, Ann1, Ann2, Ann3]
        - Instantiating an annotated type is equivalent to instantiating the
        underlying type::
            Annotated[C, Ann1](5) == C(5)
        - Annotated can be used as a generic type alias::
            Optimized = Annotated[T, runtime.Optimize()]
            Optimized[int] == Annotated[int, runtime.Optimize()]
            OptimizedList = Annotated[List[T], runtime.Optimize()]
            OptimizedList[int] == Annotated[List[int], runtime.Optimize()]
        """

        __slots__ = ()

        def __new__(cls, *args, **kwargs):
            raise TypeError("Type Annotated cannot be instantiated.")

        @_tp_cache
        def __class_getitem__(cls, params):
            if not isinstance(params, tuple) or len(params) < 2:
                raise TypeError("Annotated[...] should be used "
                                "with at least two arguments (a type and an "
                                "annotation).")
            msg = "Annotated[t, ...]: t must be a type."
            # noinspection PyUnresolvedReferences,PyProtectedMember
            origin = typing._type_check(params[0], msg)
            metadata = tuple(params[1:])
            return _AnnotatedAlias(origin, metadata)

        def __init_subclass__(cls, *args, **kwargs):
            raise TypeError(
                "Cannot subclass {}.Annotated".format(cls.__module__)
            )

    def _strip_annotations(t):  # pylint: disable=invalid-name
        """Strips the annotations from a given type.
        """
        if isinstance(t, _AnnotatedAlias):
            return _strip_annotations(t.__origin__)
        # noinspection PyUnresolvedReferences,PyProtectedMember
        if isinstance(t, typing._GenericAlias):
            stripped_args = tuple(_strip_annotations(a) for a in t.__args__)
            if stripped_args == t.__args__:
                return t
            res = t.copy_with(stripped_args)
            # noinspection PyProtectedMember
            res._special = t._special
            return res
        return t

    def get_type_hints(obj, globalns=None, localns=None, include_extras=False):
        """Return type hints for an object.
        This is often the same as obj.__annotations__, but it handles
        forward references encoded as string literals, adds Optional[t] if a
        default value equal to None is set and recursively replaces all
        'Annotated[T, ...]' with 'T' (unless 'include_extras=True').
        The argument may be a module, class, method, or function. The annotations
        are returned as a dictionary. For classes, annotations include also
        inherited members.
        TypeError is raised if the argument is not of a type that can contain
        annotations, and an empty dictionary is returned if no annotations are
        present.
        BEWARE -- the behavior of globalns and localns is counterintuitive
        (unless you are familiar with how eval() and exec() work).  The
        search order is locals first, then globals.
        - If no dict arguments are passed, an attempt is made to use the
          globals from obj (or the respective module's globals for classes),
          and these are also used as the locals.  If the object does not appear
          to have globals, an empty dictionary is used.
        - If one dict argument is passed, it is used for both globals and
          locals.
        - If two dict arguments are passed, they specify globals and
          locals, respectively.
        """
        hint = typing.get_type_hints(obj, globalns=globalns, localns=localns)
        if include_extras:
            return hint
        return {k: _strip_annotations(t) for k, t in hint.items()}

# Python 3.8 has get_origin() and get_args() but those implementations aren't
# Annotated-aware, so we can't use those, only Python 3.9 versions will do.
if sys.version_info[:2] >= (3, 9):  # pragma: no cover

    get_origin = typing.get_origin
    get_args = typing.get_args

elif PEP_560:  # pragma: no cover

    # noinspection PyUnresolvedReferences,PyProtectedMember
    from typing import _GenericAlias
    try:
        # 3.9+
        # noinspection PyProtectedMember
        from typing import _BaseGenericAlias
    except ImportError:
        _BaseGenericAlias = _GenericAlias
    try:
        # 3.9+
        from typing import GenericAlias
    except ImportError:
        GenericAlias = _GenericAlias


    def get_origin(tp):  # pylint: disable=invalid-name
        """Get the unsubscripted version of a type.

        This supports generic types, Callable, Tuple, Union, Literal, Final, ClassVar
        and Annotated. Return None for unsupported types. Examples::

            get_origin(Literal[42]) is Literal
            get_origin(int) is None
            get_origin(ClassVar[int]) is ClassVar
            get_origin(Generic) is Generic
            get_origin(Generic[T]) is Generic
            get_origin(Union[T, int]) is Union
            get_origin(List[Tuple[T, T]][int]) == list
        """
        if isinstance(tp, _AnnotatedAlias):
            return Annotated
        if isinstance(tp, (_BaseGenericAlias, GenericAlias)):
            return tp.__origin__
        if tp is typing.Generic:
            return typing.Generic
        return None


    def get_args(tp):  # pylint: disable=invalid-name
        """Get type arguments with all substitutions performed.

        For unions, basic simplifications used by Union constructor are performed.
        Examples::
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        if isinstance(tp, _AnnotatedAlias):
            return (tp.__origin__,) + tp.__metadata__
        if isinstance(tp, (_GenericAlias, GenericAlias)):
            res = tp.__args__
            if tp.__origin__ is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import datetime
import decimal
import inspect
import typing
import unittest

from sane_finances.sources import generic
from sane_finances.sources.cbr.v2016.exporters import CbrCurrencyRatesExporterFactory
from sane_finances.sources.base import (
    MaxPagesLimitExceeded, InstrumentValuesHistoryEmpty,
    InstrumentInfo, InstrumentValue, ParseError, InstrumentExporterFactory, InstrumentExporterRegistry)
from .fakes import (
    FakeInstrumentHistoryDownloadParameters, FakeInstrumentInfoProvider, FakeInstrumentValueProvider,
    FakeInstrumentInfoParser, FakeInstrumentValuesHistoryParser, FakeInstrumentStringDataDownloader,
    FakeInstrumentExporterFactory)


class TestGenericInstrumentHistoryValuesExporter(unittest.TestCase):

    def setUp(self):
        self.string_data_downloader = FakeInstrumentStringDataDownloader('FAKE_INFO', 'FAKE_HISTORY')
        self.history_values_parser = FakeInstrumentValuesHistoryParser(
            self.string_data_downloader.fake_history_string)
        self.exporter = generic.GenericInstrumentHistoryValuesExporter(
            self.string_data_downloader, self.history_values_parser)

    def _prepare_expected_result(self, moment_from: datetime.datetime, moment_to: datetime.datetime):
        # parsed values contain dates below asked interval (before moment_from and after moment_to)
        parsed_values = []
        current_date = moment_from - datetime.timedelta(days=5)
        end_date = moment_to + datetime.timedelta(days=5)
        while current_date < end_date:
            parsed_values.append(FakeInstrumentValueProvider(InstrumentValue(
                moment=current_date, value=decimal.Decimal(42))))

            current_date += datetime.timedelta(days=1)

        # but expected result must contain only dates strictly inside the asked interval
        # (note FakeInstrumentStringDataDownloader adjusted parameters)
        expected_result = [
            parsed_value
            for parsed_value
            in parsed_values
            if moment_from <= parsed_value.get_instrument_value(tzinfo=moment_from.tzinfo).moment <= moment_to]

        # assert of arrange
        self.assertGreater(len([
            value
            for value
            in parsed_values
            if value.get_instrument_value(tzinfo=moment_from.tzinfo).moment < moment_from]), 0)
        self.assertGreater(len([
            value
            for value
            in parsed_values
            if value.get_instrument_value(tzinfo=moment_from.tzinfo).moment > moment_to]), 0)
        self.assertEqual(len([
            value
            for value
            in expected_result
            if (value.get_instrument_value(tzinfo=moment_from.tzinfo).moment < moment_from
                or value.get_instrument_value(tzinfo=moment_from.tzinfo).moment > moment_to)]), 0)

        return expected_result

    def test_success(self):
        moment_from = moment_to = datetime.datetime(2000, 1, 1)

        expected_result = [FakeInstrumentValueProvider(InstrumentValue(
            moment=moment_from, value=decimal.Decimal(42)))]
        self.history_values_parser.fake_result = expected_result

        history = list(self.exporter.export_instrument_history_values(
            FakeInstrumentHistoryDownloadParameters(),
            moment_from,
            moment_to))

        self.assertSequenceEqual(history, expected_result)
        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertEqual(self.history_values_parser.parse_counter, 1)

    def test_returnOnlyAskedInterval(self):
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=5)

        # parsed values may contain dates below asked interval (before moment_from and after moment_to)
        # but expected result must contain only dates strictly inside the asked interval
        # (note that FakeInstrumentStringDataDownloader DO NOT adjust parameters here)
        expected_result = self._prepare_expected_result(moment_from, moment_to)
        self.history_values_parser.fake_result = expected_result

        history = list(self.exporter.export_instrument_history_values(
            FakeInstrumentHistoryDownloadParameters(),
            moment_from,
            moment_to))

        self.assertSequenceEqual(history, expected_result)
        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertEqual(self.history_values_parser.parse_counter, 1)

    def test_returnOnlyAdjustedInterval(self):
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=5)
        adjusted_moment_from = moment_from - datetime.timedelta(days=2)
        adjusted_moment_to = moment_to + datetime.timedelta(days=2)

        self.assertNotEqual(moment_from, adjusted_moment_from)
        self.assertNotEqual(moment_to, adjusted_moment_to)

        # parsed values may contain dates below asked interval (before moment_from and after moment_to)
        # but expected result must contain only dates strictly inside the ADJUSTED interval
        # (note that FakeInstrumentStringDataDownloader DO adjust parameters here)
        expected_result = self._prepare_expected_result(adjusted_moment_from, adjusted_moment_to)

        # imitate adjustment:
        # noinspection PyUnusedLocal,PyShadowingNames
        def fake_adjust_download_instrument_history_parameters(parameters, moment_from, moment_to):
            return parameters, adjusted_moment_from, adjusted_moment_to

        self.string_data_downloader.adjust_download_instrument_history_parameters = \
            fake_adjust_download_instrument_history_parameters

        self.history_values_parser.fake_result = expected_result

        history = list(self.exporter.export_instrument_history_values(
            FakeInstrumentHistoryDownloadParameters(),
            moment_from,
            moment_to))

        self.assertSequenceEqual(history, expected_result)
        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertEqual(self.history_values_parser.parse_counter, 1)

    def test_returnCallDownloadAndParseMultipleTimes(self):
        pages_count = 5
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=pages_count)

        # parsed values may contain dates below asked interval (before moment_from and after moment_to)
        # but expected result must contain only dates strictly inside the asked interval
        # (note that FakeInstrumentStringDataDownloader DO NOT adjust parameters here)
        parse_result = self._prepare_expected_result(moment_from, moment_to)
        expected_result = parse_result * pages_count

        # imitate pagination:
        # noinspection PyShadowingNames
        def fake_paginate_download_instrument_history_parameters(parameters, moment_from, moment_to):
            page_begin = moment_from
            while page_begin < moment_to:
                yield parameters, page_begin, page_begin
                page_begin += datetime.timedelta(days=1)

        self.string_data_downloader.paginate_download_instrument_history_parameters = \
            fake_paginate_download_instrument_history_parameters

        self.history_values_parser.fake_result = parse_result

        history = list(self.exporter.export_instrument_history_values(
            FakeInstrumentHistoryDownloadParameters(),
            moment_from,
            moment_to))

        self.assertSequenceEqual(history, expected_result)
        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, pages_count)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertEqual(self.history_values_parser.parse_counter, pages_count)

    def test_raiseWhenPagesLimitExceeded(self):
        pages_count = 5
        max_paged_parameters = pages_count - 1
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=pages_count)

        # parsed values may contain dates below asked interval (before moment_from and after moment_to)
        # but expected result must contain only dates strictly inside the asked interval
        # (note that FakeInstrumentStringDataDownloader DO NOT adjust parameters here)
        parse_result = self._prepare_expected_result(moment_from, moment_to)

        # imitate pagination:
        # noinspection PyShadowingNames
        def fake_paginate_download_instrument_history_parameters(parameters, moment_from, moment_to):
            page_begin = moment_from
            while page_begin < moment_to:
                yield parameters, page_begin, page_begin
                page_begin += datetime.timedelta(days=1)

        self.string_data_downloader.paginate_download_instrument_history_parameters = \
            fake_paginate_download_instrument_history_parameters

        self.history_values_parser.fake_result = parse_result
        self.exporter.max_paged_parameters = max_paged_parameters

        with self.assertRaises(MaxPagesLimitExceeded):
            _ = list(self.exporter.export_instrument_history_values(
                FakeInstrumentHistoryDownloadParameters(),
                moment_from,
                moment_to))

        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))

    def test_raiseWhenMomentFromGreaterThenMomentTo(self):
        self.history_values_parser.fake_result = []

        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from - datetime.timedelta(seconds=1)

        with self.assertRaises(ValueError):
            _ = list(self.exporter.export_instrument_history_values(
                FakeInstrumentHistoryDownloadParameters(),
                moment_from,
                moment_to))

    def test_acceptInstrumentValuesHistoryEmptyException(self):
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=5)

        fake_result = (result for result in ['DOESNT MATTER WHAT'])

        self.history_values_parser.parse_exception = InstrumentValuesHistoryEmpty()
        self.history_values_parser.fake_result = fake_result
        expected_result = []

        # verify arrange
        self.assertNotEqual(fake_result, expected_result)

        history = list(self.exporter.export_instrument_history_values(
            FakeInstrumentHistoryDownloadParameters(),
            moment_from,
            moment_to))

        self.assertSequenceEqual(history, expected_result)
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instrument_history_string_results))

    def test_LastDownloadStringResultIsNotCorrectWhenError(self):
        moment_from = datetime.datetime(2000, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=5)

        self.history_values_parser.parse_exception = ParseError('Error')
        self.history_values_parser.fake_result = []

        with self.assertRaises(Exception):
            _ = list(self.exporter.export_instrument_history_values(
                FakeInstrumentHistoryDownloadParameters(),
                moment_from,
                moment_to))

        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 0)
        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)


class TestGenericInstrumentsInfoExporter(unittest.TestCase):

    def setUp(self):
        self.string_data_downloader = FakeInstrumentStringDataDownloader('FAKE_INFO', 'FAKE_HISTORY')
        self.info_parser = FakeInstrumentInfoParser(self.string_data_downloader.fake_info_string)
        self.exporter = generic.GenericInstrumentsInfoExporter(
            self.string_data_downloader, self.info_parser)

    def test_success(self):
        expected_result = [
            FakeInstrumentInfoProvider(InstrumentInfo(code='SOME CODE', name='SOME NAME'))
        ]
        self.info_parser.fake_result = expected_result

        info_list = list(self.exporter.export_instruments_info(None))

        self.assertSequenceEqual(info_list, expected_result)
        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, 0)
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.string_data_downloader.download_instruments_info_string_results))
        self.assertEqual(self.info_parser.parse_counter, 1)

    def test_LastDownloadStringResultIsNotCorrectWhenError(self):
        self.info_parser.fake_result = []
        self.info_parser.parse_exception = ParseError('Error')

        with self.assertRaises(Exception):
            _ = list(self.exporter.export_instruments_info(None))

        self.assertEqual(self.string_data_downloader.download_instrument_history_string_counter, 0)
        self.assertEqual(self.string_data_downloader.download_instruments_info_string_counter, 1)
        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)


class TestRegistry(unittest.TestCase):

    def test_get_all_instrument_exporters_Success(self):
        self.assertIsNotNone(generic.get_all_instrument_exporters())

    def test_get_instrument_exporter_by_factory_SuccessByClass(self):
        factory = CbrCurrencyRatesExporterFactory
        self.assertTrue(inspect.isclass(factory))

        self.assertIsNotNone(generic.get_instrument_exporter_by_factory(factory))

    def test_get_instrument_exporter_by_factory_SuccessByInstance(self):
        factory = CbrCurrencyRatesExporterFactory()
        self.assertTrue(isinstance(factory, InstrumentExporterFactory))

        self.assertIsNotNone(generic.get_instrument_exporter_by_factory(factory))

    def test_register_instrument_history_values_exporter_Success(self):
        fake_registry = InstrumentExporterRegistry(
            name='FACTORY NAME',
            factory=FakeInstrumentExporterFactory()
        )

        generic.register_instrument_history_values_exporter(fake_registry)

    def test_register_instrument_history_values_exporter_RaiseWithWrongArguments(self):
        new_factory_class = type('NewInstrumentExporterFactory', (FakeInstrumentExporterFactory,), {})

        expected_exceptions: typing.Tuple = (TypeError, ValueError)

        # can't register None
        with self.assertRaises(expected_exceptions):
            # noinspection PyTypeChecker
            generic.register_instrument_history_values_exporter(None)

        # can't register wrong type
        with self.assertRaises(expected_exceptions):
            # noinspection PyTypeChecker
            generic.register_instrument_history_values_exporter(object())

        # can't register with no name
        with self.assertRaises(expected_exceptions):
            # noinspection PyTypeChecker
            generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
                name=None,
                factory=new_factory_class()))

        # can't register with empty name
        with self.assertRaises(expected_exceptions):
            generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
                name='      ',  # spaces
                factory=new_factory_class()))

        # can't register with no factory
        with self.assertRaises(expected_exceptions):
            # noinspection PyTypeChecker
            generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
                name='FACTORY NAME',
                factory=None))

        # can't register with wrong factory
        with self.assertRaises(expected_exceptions):
            # noinspection PyTypeChecker
            generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
                name='FACTORY NAME',
                factory=object()))

        # register some factory
        generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
            name='FACTORY NAME',
            factory=new_factory_class()))

        # can't register same factory class
        with self.assertRaises(expected_exceptions):
            generic.register_instrument_history_values_exporter(InstrumentExporterRegistry(
                name='ANOTHER FACTORY NAME',
                factory=new_factory_class()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import ParseError, CheckApiActualityError, InstrumentExporterFactory
from sane_finances.sources.solactive.v2018.exporters import (
    SolactiveStringDataDownloader, SolactiveApiActualityChecker, SolactiveExporterFactory,
    SolactiveDownloadParameterValuesStorage)
from sane_finances.sources.solactive.v2018.meta import (
    IndexValue, SolactiveIndexHistoryDownloadParameters, SolactiveIndexesInfoDownloadParameters, IndexHistoryTypes,
    FileExtensions, FieldNames)
from .common import CommonTestCases
from .fakes import (
    FakeDownloader, FakeSolactiveStringDataDownloader, FakeSolactiveJsonParser)


class TestSolactiveStringDataDownloader(unittest.TestCase):
    
    def setUp(self):
        # fake downloaders
        self.fake_data = 'data'
        
        downloader = FakeDownloader(self.fake_data)
        
        self.string_data_downloader = SolactiveStringDataDownloader(downloader)

    def test_paginate_download_instrument_history_parameters(self):
        isin = 'SOME_ISIN'
        parameters = SolactiveIndexHistoryDownloadParameters.safe_create(isin=isin)
        moment_from, moment_to = datetime.datetime(2010, 1, 1), datetime.datetime(2020, 1, 1)

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertTrue(len(paginated_parameters_tuples) >= 1)

    def test_download_instrument_history_string(self):
        isin = 'SOME_ISIN'
        parameters = SolactiveIndexHistoryDownloadParameters.safe_create(isin=isin)
        moment_from = moment_to = datetime.datetime(2000, 1, 1)

        result = self.string_data_downloader.download_instrument_history_string(parameters, moment_from, moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instruments_info_string(self):
        parameters = SolactiveIndexesInfoDownloadParameters.safe_create()

        result = self.string_data_downloader.download_instruments_info_string(parameters)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_index_history_string(self):
        isin = 'SOME_ISIN'
        
        result = self.string_data_downloader.download_index_history_string(isin)
        
        self.assertEqual(result.downloaded_string, self.fake_data)
    
    def test_download_all_indexes_info_success(self):
        result = self.string_data_downloader.download_all_indexes_info_string()
        
        self.assertEqual(result.downloaded_string, self.fake_data)


class TestSolactiveAPIActualityChecker(unittest.TestCase):
    
    def setUp(self):
        # fakes
        self.success_json_parsed_item = \
            IndexValue(
                index_id=SolactiveApiActualityChecker.ActualityCheckISIN,
                # convert to datetime:
                moment=datetime.datetime.combine(SolactiveApiActualityChecker._expectedFirstMomentDate,
                                                 datetime.time.min),
                value=SolactiveApiActualityChecker._expectedFirstValue
            )
        
        self.fake_string_data_downloader = FakeSolactiveStringDataDownloader(None, None)
        self.json_parser = FakeSolactiveJsonParser([self.success_json_parsed_item])
        self.checker = SolactiveApiActualityChecker(self.fake_string_data_downloader, self.json_parser)

    def test_check_Success(self):
        checker = self.checker
        
        checker.check()

        # check that there is no incorrect downloaded strings
        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.fake_string_data_downloader.download_instrument_history_string_results))

    def test_check_raisesWhenHistoryParseError(self):
        # corrupt data
        self.json_parser.parse_exception = ParseError('Error')
        checker = self.checker

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryUnknownError(self):
        # corrupt data
        self.json_parser.parse_exception = Exception('Error')
        checker = self.checker

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongISIN(self):
        # corrupt data
        self.success_json_parsed_item.index_id = 'WRONG'
        checker = self.checker
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongMoment(self):
        # corrupt data
        self.success_json_parsed_item.moment = datetime.datetime.max
        checker = self.checker
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongValue(self):
        # corrupt data
        self.success_json_parsed_item.value = decimal.Decimal(-1)
        checker = self.checker
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()


class TestSolactiveDownloadParameterValuesStorage(unittest.TestCase):

    def setUp(self):
        self.storage = SolactiveDownloadParameterValuesStorage()

    def test_is_dynamic_enum_type_AlwaysFalse(self):
        self.assertFalse(self.storage.is_dynamic_enum_type(IndexHistoryTypes))
        self.assertFalse(self.storage.is_dynamic_enum_type(FileExtensions))
        self.assertFalse(self.storage.is_dynamic_enum_type(FieldNames))
        # noinspection PyTypeChecker
        self.assertFalse(self.storage.is_dynamic_enum_type(None))

    def test_get_dynamic_enum_key_AlwaysNone(self):
        self.assertIsNone(self.storage.get_dynamic_enum_key(IndexHistoryTypes.MAX))
        self.assertFalse(self.storage.get_dynamic_enum_key(FileExtensions.JSON))
        self.assertFalse(self.storage.get_dynamic_enum_key(FieldNames.INDEX_ID))

    def test_get_all_managed_types_AlwaysEmpty(self):
        self.assertSequenceEqual(list(self.storage.get_all_managed_types()), [])


class TestMsciIndexExporterFactory(CommonTestCases.CommonIndexExporterFactoryTests):

    def get_exporter_factory(self) -> InstrumentExporterFactory:
        return SolactiveExporterFactory()

    def is_dynamic_enum_type_manager_singleton(self):
        return True

    def is_download_parameters_factory_singleton(self):
        return True



#!/usr/bin/python
# -*- coding: utf-8 -*-

import inspect
import unittest

from sane_finances.sources.base import (
    InstrumentExporterFactory, DownloadParametersFactory, InstrumentHistoryDownloadParameters)
from ..communication.fakes import FakeDownloader


class CommonTestCases:  # hide test cases from unittest discovery

    class CommonDownloadParametersFactoryTests(unittest.TestCase):

        def get_download_parameters_factory(self) -> DownloadParametersFactory:
            raise NotImplementedError

        def get_generate_history_download_parameters_from_expected_result_with_none(self) \
                -> InstrumentHistoryDownloadParameters:
            raise NotImplementedError

        def test_download_history_parameters_class_Success(self):
            factory = self.get_download_parameters_factory()

            self.assertTrue(inspect.isclass(factory.download_history_parameters_class))

        def test_download_history_parameters_factory_Success(self):
            factory = self.get_download_parameters_factory()

            self.assertTrue(callable(factory.download_history_parameters_factory))

        def test_download_info_parameters_class_Success(self):
            factory = self.get_download_parameters_factory()

            self.assertTrue(inspect.isclass(factory.download_info_parameters_class))

        def test_download_info_parameters_factory_Success(self):
            factory = self.get_download_parameters_factory()

            self.assertTrue(callable(factory.download_info_parameters_factory))

        def test_generate_history_download_parameters_from_SuccessWithNone(self):
            factory = self.get_download_parameters_factory()
            expected_result = self.get_generate_history_download_parameters_from_expected_result_with_none()

            result = factory.generate_history_download_parameters_from(None, None, None)

            self.assertEqual(result, expected_result)

    class CommonIndexExporterFactoryTests(unittest.TestCase):

        def get_exporter_factory(self) -> InstrumentExporterFactory:
            raise NotImplementedError

        def is_dynamic_enum_type_manager_singleton(self):
            raise NotImplementedError

        def is_download_parameters_factory_singleton(self):
            raise NotImplementedError

        def test_create_history_values_exporter_Success(self):
            factory = self.get_exporter_factory()
            result = factory.create_history_values_exporter(FakeDownloader(None))

            self.assertIsNotNone(result)

        def test_create_info_exporter_Success(self):
            factory = self.get_exporter_factory()
            result = factory.create_info_exporter(FakeDownloader(None))

            self.assertIsNotNone(result)

        def test_create_download_parameter_values_storage_Success(self):
            factory = self.get_exporter_factory()
            result = factory.create_download_parameter_values_storage(FakeDownloader(None))

            self.assertIsNotNone(result)

        def test_create_api_actuality_checker_Success(self):
            factory = self.get_exporter_factory()
            result = factory.create_api_actuality_checker(FakeDownloader(None))

            self.assertIsNotNone(result)

        def test_dynamic_enum_type_manager_Success(self):
            factory = self.get_exporter_factory()
            result = factory.dynamic_enum_type_manager

            self.assertIsNotNone(result)

            if self.is_dynamic_enum_type_manager_singleton():
                new_result = factory.dynamic_enum_type_manager

                self.assertIs(new_result, result)

        def test_download_parameters_factory_Success(self):
            factory = self.get_exporter_factory()
            result = factory.download_parameters_factory

            self.assertIsNotNone(result)

            if self.is_download_parameters_factory_singleton():
                new_result = factory.download_parameters_factory

                self.assertIs(new_result, result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
import datetime

from sane_finances.communication.downloader import DownloadStringResult
from sane_finances.sources.solactive.v2018.meta import (
    IndexValue, IndexInfo)
from sane_finances.sources.solactive.v2018.parsers import (
    SolactiveJsonParser, SolactiveIndexInfoParser)
from sane_finances.sources.solactive.v2018.exporters import (
    SolactiveStringDataDownloader)

from ....communication.fakes import FakeDownloader


class FakeSolactiveStringDataDownloader(SolactiveStringDataDownloader):

    def __init__(self, fake_info_data, fake_history_data):
        super().__init__(FakeDownloader(None))

        self.download_instruments_info_string_results: typing.List[DownloadStringResult] = []
        self.download_instrument_history_string_results: typing.List[DownloadStringResult] = []

        self.fake_history_data = fake_history_data
        self.fake_info_data = fake_info_data

    def download_index_history_string(self, isin):
        result = DownloadStringResult(self.fake_history_data)
        self.download_instrument_history_string_results.append(result)
        return result

    def download_all_indexes_info_string(self):
        result = DownloadStringResult(self.fake_info_data)
        self.download_instruments_info_string_results.append(result)
        return result


class FakeSolactiveJsonParser(SolactiveJsonParser):

    def __init__(self, fake_data: typing.Iterable[IndexValue]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(self, raw_json_text: str, tzinfo: typing.Optional[datetime.timezone]) -> typing.Iterable[IndexValue]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeSolactiveIndexInfoParser(SolactiveIndexInfoParser):

    def __init__(self, fake_data: typing.Iterable[IndexInfo]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(self, html) -> typing.Iterable[IndexInfo]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Abstract classes for data downloaders (readers).
"""

import abc
import typing


class DownloadError(Exception):
    """ Base exception for all exceptions generated by downloaders. """
    pass


class DownloadStringResult:
    """ Contains downloaded string
    and provides mechanism for feedback of downloaded string quality (actuality, correctness, fullness, etc.)
    to downloader (e.g. for cache and throttle management).
    """

    def __init__(self, downloaded_string: str):
        """ Initialize instance.

        :param downloaded_string: Value of downloaded string.
        """
        self._downloaded_string = downloaded_string
        self._is_correct = None

    @property
    def downloaded_string(self):
        """ Stored downloaded string.
        """
        return self._downloaded_string

    @property
    def is_correct(self) -> typing.Optional[bool]:
        """ Correctness of stored downloaded string.

        :return: ``True`` - downloaded string was parsed and verified successfully;
            ``False`` - downloaded string was parsed or verified as incorrect;
            ``None`` - correctness os downloaded string is not known yet.
        """
        return self._is_correct

    def set_correctness(self, is_correct: bool):
        """ Change downloaded string correctness.
        Subclasses may override this method for feedback management of specific downloader.

        :param is_correct: New value of correctness.
            Must be ``True`` or ``False``. ``None`` interprets as ``False``.
        """
        self._is_correct = bool(is_correct)


class Downloader(abc.ABC):
    """ Base abstract downloader.
    """
    @property
    @abc.abstractmethod
    def parameters(self) -> typing.List[typing.Tuple[str, str]]:
        """ Current value of download request parameters: ``[(param_code, param_value)]``
        """
        raise NotImplementedError

    @parameters.setter
    @abc.abstractmethod
    def parameters(self, value: typing.List[typing.Tuple[str, str]]) -> None:
        """ Set download request parameters value.

        :param value: New download request parameters value.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def headers(self) -> typing.Dict[str, str]:
        """ Current value of download request headers: ``{header_code : header_value}``
        """
        raise NotImplementedError

    @headers.setter
    @abc.abstractmethod
    def headers(self, value: typing.Dict[str, str]) -> None:
        """ Set download request headers value.

        :param value: New download request headers value.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def download_string(self, url: str, encoding: str = None) -> DownloadStringResult:
        """ Downloads string from `url`.

        Raise ``DownloadError`` on errors.

        :param url: URL for download
        :param encoding: Encoding of downloaded string.
            If ``None`` uses default encoding (depends of concrete downloader).
        :return: Container with downloaded string.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Metadata for export index data from moex.com
"""

import decimal
import typing
import enum
import datetime
import dataclasses

from ...base import (
    InstrumentValue, InstrumentInfo, InstrumentValueProvider, InstrumentInfoProvider,
    InstrumentHistoryDownloadParameters, DownloadParametersFactory)
from ...inspection import InstrumentInfoParameter
from ....annotations import LEGACY_ANNOTATIONS, Volatile

if LEGACY_ANNOTATIONS:  # pragma: no cover
    from ....annotations import Annotated
else:  # pragma: no cover
    from typing import Annotated


class ResponseFormats(enum.Enum):
    """ Response formats
    """
    XML = 'xml'
    CSV = 'csv'
    JSON = 'json'
    HTML = 'html'


class JsonFormats(enum.Enum):
    """ JSON formats
    """
    COMPACT = 'compact'
    EXTENDED = 'extended'


class Limits(enum.Enum):
    """ Limits of data (page size) in one response
    """
    ONE = 1
    FIVE = 5
    TEN = 10
    TWENTY = 20
    FIFTY = 50
    HUNDRED = 100


class TradeEngine(typing.NamedTuple):
    """ Trade engine from moex.com.
    """
    identity: int
    name: str
    title: str

    @classmethod
    def safe_create(
            cls: typing.Type['TradeEngine'],
            *,
            identity: int,
            name: str,
            title: str) -> 'TradeEngine':
        """ Create new instance of ``TradeEngine`` with arguments check.

        :param identity: Identity value.
        :param name: Name.
        :param title: Title.
        :return: ``TradeEngine`` instance.
        """
        return cls(identity=int(identity), name=str(name), title=str(title))


class Market(typing.NamedTuple):
    """ Market from moex.com.
    """
    identity: int
    trade_engine: TradeEngine
    name: str
    title: str
    marketplace: str

    @classmethod
    def safe_create(
            cls: typing.Type['Market'],
            *,
            identity: int,
            trade_engine: TradeEngine,
            name: str,
            title: str,
            marketplace: str) -> 'Market':
        """ Create new instance of ``Market`` with arguments check.

        :param identity: Identity value.
        :param trade_engine: Trade engine.
        :param name: Name.
        :param title: Title.
        :param marketplace: Marketplace.
        :return: ``Market`` instance.
        """
        if not isinstance(trade_engine, TradeEngine):
            raise TypeError("'trade_engine' is not TradeEngine")

        return cls(
            identity=int(identity),
            trade_engine=trade_engine,
            name=str(name),
            title=str(title),
            marketplace=str(marketplace))


class Board(typing.NamedTuple):
    """ Board from moex.com.
    """
    identity: int
    trade_engine: TradeEngine
    market: Market
    boardid: str
    title: str
    is_traded: bool
    has_candles: bool
    is_primary: bool

    @classmethod
    def safe_create(
            cls: typing.Type['Board'],
            *,
            identity: int,
            trade_engine: TradeEngine,
            market: Market,
            boardid: str,
            title: str,
            is_traded: bool,
            has_candles: bool,
            is_primary: bool) -> 'Board':
        """ Create new instance of ``Board`` with arguments check.

        :param identity: Identity value.
        :param trade_engine: Trade engine.
        :param market: Market.
        :param boardid: Board ID.
        :param title: Title.
        :param is_traded: Is board traded.
        :param has_candles: Has board candles.
        :param is_primary: Is board primary.
        :return: ``Board`` instance.
        """
        if not isinstance(trade_engine, TradeEngine):
            raise TypeError("'trade_engine' is not TradeEngine")
        if not isinstance(market, Market):
            raise TypeError("'market' is not Market")

        return cls(
            identity=int(identity),
            trade_engine=trade_engine,
            market=market,
            boardid=str(boardid),
            title=str(title),
            is_traded=bool(is_traded),
            has_candles=bool(has_candles),
            is_primary=bool(is_primary))


class GlobalIndexData(typing.NamedTuple):
    """ Container for global index data from moex.com.
    """
    trade_engines: typing.Tuple[TradeEngine, ...]
    markets: typing.Tuple[Market, ...]
    boards: typing.Tuple[Board, ...]


@dataclasses.dataclass
class SecurityValue(InstrumentValueProvider):
    """ Container for security history item.
    """
    trade_date: datetime.date
    close: decimal.Decimal

    def __init__(self,
                 *,
                 trade_date: datetime.date,
                 close: decimal.Decimal):
        """ Initialize instance.

        :param trade_date: Trade date.
        :param close: Close value.
        """
        if not isinstance(trade_date, datetime.date):
            raise TypeError("'trade_date' is not date")

        self.trade_date = trade_date
        self.close = decimal.Decimal(close)

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        return InstrumentValue(
            value=self.close,
            moment=datetime.datetime.combine(self.trade_date, datetime.time.min, tzinfo=tzinfo))


@dataclasses.dataclass
class SecurityInfo(InstrumentInfoProvider):
    """ Container for security information.
    """
    sec_id: str
    board: Board
    short_name: str
    lot_size: typing.Optional[int]
    sec_name: typing.Optional[str]
    isin: typing.Optional[str]
    lat_name: typing.Optional[str]
    reg_number: typing.Optional[str]
    coupon_period: typing.Optional[int]
    coupon_percent: typing.Optional[float]

    def __init__(self,
                 *,
                 sec_id: str,
                 board: Board,
                 short_name: str,
                 lot_size: int = None,
                 sec_name: str = None,
                 isin: str = None,
                 lat_name: str = None,
                 reg_number: str = None,
                 coupon_period: int = None,
                 coupon_percent: float = None):
        if not isinstance(board, Board):
            raise TypeError("'board' is not Board")

        self.sec_id = str(sec_id)
        self.board = board
        self.short_name = str(short_name)
        self.lot_size = None if lot_size is None else int(lot_size)
        self.sec_name = None if sec_name is None else str(sec_name)
        self.isin = None if isin is None else str(isin)
        self.lat_name = None if lat_name is None else str(lat_name)
        self.reg_number = None if reg_number is None else str(reg_number)
        self.coupon_period = None if coupon_period is None else int(coupon_period)
        self.coupon_percent = None if coupon_percent is None else float(coupon_percent)

    def __str__(self):
        return (f"MOEX security ("
                f"sec_id={self.sec_id!r}, "
                f"short_name={self.short_name!r}, "
                f"sec_name={self.sec_name!r}, "
                f"isin={self.isin!r}, "
                f"lat_name={self.lat_name!r}, "
                f"lot_size={self.lot_size}, "
                f"engine={self.board.trade_engine.name!r}, "
                f"market={self.board.market.name!r}, "
                f"board={self.board.boardid!r}, "
                f"reg_number={self.reg_number!r}, "
                f"coupon_period={self.coupon_period}, "
                f"coupon_percent={self.coupon_percent})")

    @property
    def instrument_info(self) -> InstrumentInfo:
        return InstrumentInfo(code=self.sec_id, name=self.short_name)


class MoexSecuritiesInfoDownloadParameters(typing.NamedTuple):
    """ Container for ``MoexStringDataDownloader.download_instruments_info_string`` parameters.
    """
    board: Board

    @classmethod
    def safe_create(
            cls: typing.Type['MoexSecuritiesInfoDownloadParameters'],
            *,
            board: Board) -> 'MoexSecuritiesInfoDownloadParameters':
        """ Create new instance of ``MoexSecuritiesInfoDownloadParameters`` with arguments check.

        :param board: Board.
        :return: ``MoexSecuritiesInfoDownloadParameters`` instance.
        """
        if not isinstance(board, Board):
            raise TypeError("'board' is not Board")

        return cls(board=board)


@dataclasses.dataclass
class MoexSecurityHistoryDownloadParameters(InstrumentHistoryDownloadParameters):
    """ Container for ``MoexStringDataDownloader.download_instrument_history_string`` parameters.
    """
    board: Annotated[Board, InstrumentInfoParameter()]
    sec_id: Annotated[str, InstrumentInfoParameter(instrument_identity=True)]
    start: Annotated[int, Volatile(generator=lambda ctx: 0, stub_value=0)]

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Optional[MoexSecuritiesInfoDownloadParameters],
            instrument_info: typing.Optional[SecurityInfo]) -> 'MoexSecurityHistoryDownloadParameters':
        return MoexSecurityHistoryDownloadParameters.generate_from(self, info_download_parameters, instrument_info)

    @classmethod
    def generate_from(
            cls: typing.Type['MoexSecurityHistoryDownloadParameters'],
            history_download_parameters: typing.Optional['MoexSecurityHistoryDownloadParameters'],
            info_download_parameters: typing.Optional[MoexSecuritiesInfoDownloadParameters],
            instrument_info: typing.Optional[SecurityInfo]) -> 'MoexSecurityHistoryDownloadParameters':
        """ Create new history download parameters instance with data from its arguments.

        :param history_download_parameters: Optional instrument history download parameters for cloning.
        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
            `info_download_parameters` and `instrument_info`.
        """
        return cls(
            board=((None if history_download_parameters is None else history_download_parameters.board)
                   if info_download_parameters is None
                   else info_download_parameters.board),
            sec_id=((None if history_download_parameters is None else history_download_parameters.sec_id)
                    if instrument_info is None
                    else instrument_info.sec_id),
            start=(0 if history_download_parameters is None else history_download_parameters.start)
        )

    @classmethod
    def safe_create(
            cls: typing.Type['MoexSecurityHistoryDownloadParameters'],
            *,
            board: Board,
            sec_id: str,
            start: int) -> 'MoexSecurityHistoryDownloadParameters':
        """ Create new instance of ``MoexSecurityHistoryDownloadParameters`` with arguments check.

        :param board: Board.
        :param sec_id: Security ID.
        :param start: Start value.
        :return: ``MoexSecurityHistoryDownloadParameters`` instance.
        """
        if not isinstance(board, Board):
            raise TypeError(f"{board!r} is not Board")

        return cls(
            board=board,
            sec_id=str(sec_id),
            start=int(start))


class MoexDownloadParametersFactory(DownloadParametersFactory):
    """ Download parameters factories and generators for Moscow Exchange.
    """

    @property
    def download_history_parameters_class(self) -> typing.Type[MoexSecurityHistoryDownloadParameters]:
        return MoexSecurityHistoryDownloadParameters

    @property
    def download_history_parameters_factory(self) -> typing.Callable[..., MoexSecurityHistoryDownloadParameters]:
        return MoexSecurityHistoryDownloadParameters.safe_create

    @property
    def download_info_parameters_class(self):
        return MoexSecuritiesInfoDownloadParameters

    @property
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        return MoexSecuritiesInfoDownloadParameters.safe_create

    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[MoexSecurityHistoryDownloadParameters],
            info_download_parameters: typing.Optional[MoexSecuritiesInfoDownloadParameters],
            instrument_info: typing.Optional[SecurityInfo]) -> MoexSecurityHistoryDownloadParameters:
        return MoexSecurityHistoryDownloadParameters.generate_from(
            history_download_parameters,
            info_download_parameters,
            instrument_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utilities for parse data from app2.msci.com
"""

import decimal
import logging
import typing
from xml.etree import ElementTree
import datetime

from .meta import IndexValue, IndexInfo, Styles, Sizes
from ...base import InstrumentValuesHistoryParser, InstrumentInfoParser, ParseError


logging.getLogger().addHandler(logging.NullHandler())


class MsciHistoryXmlParser(InstrumentValuesHistoryParser):
    """ Parser for history data of index from XML string.

    E.g.::

      <?xml version="1.0" ?>
      <performance>
        <index id="EAFE,C,36">
          <asOf>
            <date>03/11/2016</date>
            <value>1,644.941</value>
          </asOf>
          ...
        </index>
      </performance>
    """
    RootTag = 'performance'
    date_format = '%m/%d/%Y'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(  # pylint: disable=arguments-renamed
            self,
            raw_xml_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[IndexValue]:

        try:
            root = ElementTree.fromstring(raw_xml_text)
        except ElementTree.ParseError as ex:
            raise ParseError(ex.msg) from ex

        expected_tag = self.RootTag
        if root.tag != expected_tag:
            raise ParseError(f"Wrong XML format. Root ('{root.tag}') is not '{expected_tag}'.")

        has_any = False
        for index_element in root.iterfind('./index'):
            str_id: str = index_element.attrib['id']

            self.logger.debug(f"Got index '{str_id}'")

            index_id_parts = str_id.split(',')
            if len(index_id_parts) != 3:
                raise ParseError(f"Wrong XML format. Unexpected index ID: '{str_id}'.")

            index_name, index_style, index_size = index_id_parts

            # see https://github.com/PyCQA/pylint/issues/1801 for pylint disable hint details
            index_style, index_size = Styles(index_style), Sizes(index_size)  # pylint: disable=no-value-for-parameter

            for data_element in index_element.iterfind('./asOf'):

                # get last item
                date_raw_text = None
                for date_element in data_element.iterfind('./date'):
                    date_raw_text = date_element.text

                # get last item
                value_raw_text = None
                for value in data_element.iterfind('./value'):
                    value_raw_text = value.text

                self.logger.debug(f"Got {date_raw_text!r} -> {value_raw_text!r}")

                if date_raw_text is None:
                    raise ParseError(f"Wrong XML format. Not found date tag in\n{ElementTree.tostring(data_element)}")
                if value_raw_text is None:
                    raise ParseError(f"Wrong XML format. Not found value tag in\n{ElementTree.tostring(data_element)}")

                try:
                    value_date = datetime.datetime.strptime(date_raw_text, self.date_format)
                except (ValueError, TypeError) as ex:
                    raise ParseError(f"Wrong XML format."
                                     f"Not valid date: {date_raw_text!r}") from ex

                value_date = value_date.date()

                try:
                    value = decimal.Decimal(value_raw_text.replace(',', '').replace(' ', ''))
                except (ValueError, TypeError, decimal.DecimalException) as ex:
                    raise ParseError(f"Wrong XML format."
                                     f"Not valid value: {value_raw_text!r}") from ex

                has_any = True
                yield IndexValue(
                    date=value_date,
                    value=value,
                    index_name=index_name,
                    style=index_style,
                    size=index_size)

        if not has_any:
            # empty sequence make no sense: there always must be history
            raise ParseError("Wrong XML format. Data not found.")


class MsciIndexInfoParser(InstrumentInfoParser):
    """ Parser for indexes info list from XML.

    E.g.::

        <?xml version="1.0" ?><indices><index id="2670" name="AC AMERICAS" />...</indices>
    """

    RootTag = 'indices'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(self, raw_xml_text: str) -> typing.Iterable[IndexInfo]:  # pylint: disable=arguments-renamed
        raw_xml_text = raw_xml_text.replace('&', '&amp;')  # yes, it can contain symbol & inside 'name' attribute
        try:
            root = ElementTree.fromstring(raw_xml_text)
        except ElementTree.ParseError as ex:
            raise ParseError(ex.msg) from ex

        expected_tag = self.RootTag
        if root.tag != expected_tag:
            raise ParseError(f"Wrong XML format. Root ('{root.tag}') is not '{expected_tag}'.")

        id_tag = 'id'
        name_tag = 'name'

        has_any = False
        for data_element in root.iterfind('./index'):

            if id_tag not in data_element.attrib:
                self.logger.error(f"Index ID not found in\n{ElementTree.tostring(data_element)}")
                raise ParseError("Unexpected XML format. Index ID not found")

            if name_tag not in data_element.attrib:
                self.logger.error(f"Index name not found in\n{ElementTree.tostring(data_element)}")
                raise ParseError("Unexpected XML format. Index name not found")

            index_id = data_element.attrib[id_tag]
            index_name = data_element.attrib[name_tag]

            has_any = True
            yield IndexInfo(index_id=index_id, name=index_name)

        if not has_any:
            self.logger.error(f"No data found in XML:\n{raw_xml_text}")
            raise ParseError("Unexpected XML format. No data found")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import (
    InstrumentValue, InstrumentInfo, DownloadParametersFactory, InstrumentHistoryDownloadParameters)
from sane_finances.sources.msci.v2021.meta import (
    Market, Currency, IndexLevel, IndexSuite, Size, Style, Scopes,
    IndexValue, IndexInfo, MsciIndexHistoryDownloadParameters, MsciIndexesInfoDownloadParameters,
    MsciDownloadParametersFactory)
from .common import CommonTestCases


class TestMarket(unittest.TestCase):

    def test_safe_create_raiseWrongScope(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = Market.safe_create(identity='ID', name='NAME', scope=42)


class TestIndexSuite(unittest.TestCase):

    def test_safe_create_raiseWrongScope(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexSuite.safe_create(identity='ID', name='NAME', group=42)


class TestIndexValue(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_value = IndexValue(
            calc_date=datetime.date(2000, 12, 31),
            level_eod=decimal.Decimal(42),
            msci_index_code='990300',
            index_variant_type=IndexLevel(identity='ID', name='NAME'),
            currency=Currency(identity='ID', name='NAME'))
        expected_instrument_value = InstrumentValue(
            value=index_value.level_eod,
            moment=datetime.datetime.combine(index_value.calc_date, datetime.time.min, tzinfo=datetime.timezone.utc))

        instrument_value = index_value.get_instrument_value(datetime.timezone.utc)

        self.assertEqual(expected_instrument_value, instrument_value)

    def test_raiseWrongDate(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexValue(
                calc_date=None,
                level_eod=decimal.Decimal(42),
                msci_index_code='990300',
                index_variant_type=IndexLevel(identity='ID', name='NAME'),
                currency=Currency(identity='ID', name='NAME'))

    def test_raiseWrongLevel(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexValue(
                calc_date=datetime.date(2000, 12, 31),
                level_eod=decimal.Decimal(42),
                msci_index_code='990300',
                index_variant_type=None,
                currency=Currency(identity='ID', name='NAME'))

    def test_raiseWrongCurrency(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexValue(
                calc_date=datetime.date(2000, 12, 31),
                level_eod=decimal.Decimal(42),
                msci_index_code='990300',
                index_variant_type=IndexLevel(identity='ID', name='NAME'),
                currency=None)


class TestIndexInfo(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_info = IndexInfo(
            msci_index_code='CODE',
            index_name='NAME')
        expected_instrument_info = InstrumentInfo(code=index_info.msci_index_code, name=index_info.index_name)

        instrument_info = index_info.instrument_info

        self.assertEqual(expected_instrument_info, instrument_info)


class TestMsciIndexesInfoDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = MsciIndexesInfoDownloadParameters.safe_create(
            index_scope=Scopes.REGIONAL,
            index_market=Market(identity='ID', name='NAME'),
            index_size=Size(identity='ID', name='NAME'),
            index_style=Style(identity='ID', name='NAME'),
            index_suite=IndexSuite(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongScope(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                index_scope=None,
                index_market=Market(identity='ID', name='NAME'),
                index_size=Size(identity='ID', name='NAME'),
                index_style=Style(identity='ID', name='NAME'),
                index_suite=IndexSuite(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongMarket(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                index_scope=Scopes.REGIONAL,
                index_market=None,
                index_size=Size(identity='ID', name='NAME'),
                index_style=Style(identity='ID', name='NAME'),
                index_suite=IndexSuite(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongSize(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                index_scope=Scopes.REGIONAL,
                index_market=Market(identity='ID', name='NAME'),
                index_size=None,
                index_style=Style(identity='ID', name='NAME'),
                index_suite=IndexSuite(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongStyle(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                index_scope=Scopes.REGIONAL,
                index_market=Market(identity='ID', name='NAME'),
                index_size=Size(identity='ID', name='NAME'),
                index_style=None,
                index_suite=IndexSuite(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongIndexSuite(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                index_scope=Scopes.REGIONAL,
                index_market=Market(identity='ID', name='NAME'),
                index_size=Size(identity='ID', name='NAME'),
                index_style=Style(identity='ID', name='NAME'),
                index_suite=None)


class TestMsciIndexHistoryDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = MsciIndexHistoryDownloadParameters.safe_create(
            index_code='CODE',
            currency=Currency(identity='ID', name='NAME'),
            index_variant=IndexLevel(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongCurrency(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_code='CODE',
                currency=None,
                index_variant=IndexLevel(identity='ID', name='NAME'))

    def test_safe_create_raiseWrongIndexLevel(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_code='CODE',
                currency=Currency(identity='ID', name='NAME'),
                index_variant=None)

    def test_clone_with_instrument_info_parameters_Success(self):
        params = MsciIndexHistoryDownloadParameters(
            index_code='CODE',
            currency=Currency(identity='ID', name='NAME'),
            index_variant=IndexLevel(identity='ID', name='NAME'))
        expected_result = params

        result = params.clone_with_instrument_info_parameters(None, None)

        self.assertEqual(expected_result, result)


class TestMsciDownloadParametersFactory(CommonTestCases.CommonDownloadParametersFactoryTests):

    def get_download_parameters_factory(self) -> DownloadParametersFactory:
        return MsciDownloadParametersFactory()

    def get_generate_history_download_parameters_from_expected_result_with_none(self) \
            -> InstrumentHistoryDownloadParameters:
        # noinspection PyTypeChecker
        expected_result = MsciIndexHistoryDownloadParameters(index_code=None, currency=None, index_variant=None)
        return expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import CheckApiActualityError, InstrumentExporterFactory, ParseError
from sane_finances.sources.msci.v1.exporters import (
    MsciStringDataDownloader, MsciApiActualityChecker, MsciIndexExporterFactory,
    MsciIndexDownloadParameterValuesStorage)
from sane_finances.sources.msci.v1.meta import (
    IndexValue, IndexInfo, Markets, Styles, Sizes, Scopes, Context, IndexLevels, Currencies,
    MsciIndexesInfoDownloadParameters, MsciIndexHistoryDownloadParameters)
from .common import CommonTestCases
from .fakes import (
    FakeDownloader, FakeIndexInfoParser, FakeMsciHistoryXmlParser, FakeMsciStringDataDownloader)


class TestMsciStringDataDownloader(unittest.TestCase):
    
    def setUp(self):
        # fake downloaders
        self.fake_data = 'data'
        
        downloader = FakeDownloader(self.fake_data)
        
        self.string_data_downloader = MsciStringDataDownloader(downloader)

    @staticmethod
    def generate_history_download_parameters(date_from: datetime.date, date_to: datetime.date):
        index_id = '100'
        context = Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL)
        index_level = IndexLevels.PRICE
        currency = Currencies.USD
        parameters = MsciIndexHistoryDownloadParameters.safe_create(
            index_id=index_id,
            context=context,
            index_level=index_level,
            currency=currency,
            date_from=date_from,
            date_to=date_to
        )
        return parameters

    def test_adjust_download_instrument_history_parameters_Success(self):
        moment_from = datetime.datetime(2010, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        params, moment_from, moment_to = self.string_data_downloader.adjust_download_instrument_history_parameters(
            parameters=parameters,
            moment_from=moment_from,
            moment_to=moment_to)

        self.assertLessEqual(moment_from, moment_to)
        self.assertIsNotNone(params)

        # disalign dates with parameters
        moment_from -= datetime.timedelta(days=30)
        moment_to += datetime.timedelta(days=30)

        params, moment_from, moment_to = self.string_data_downloader.adjust_download_instrument_history_parameters(
            parameters=parameters,
            moment_from=moment_from,
            moment_to=moment_to)

        self.assertLessEqual(moment_from, moment_to)
        self.assertIsNotNone(params)

        # disalign dates with parameters
        moment_from += datetime.timedelta(days=60)
        moment_to += datetime.timedelta(days=60)

        params, moment_from, moment_to = self.string_data_downloader.adjust_download_instrument_history_parameters(
            parameters=parameters,
            moment_from=moment_from,
            moment_to=moment_to)

        self.assertLessEqual(moment_from, moment_to)
        self.assertIsNotNone(params)

    def test_paginate_download_instrument_history_parameters_LongAgoSuccess(self):
        # long ago
        moment_from = datetime.datetime(1955, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 1)  # don't split

    def test_paginate_download_instrument_history_parameters_TodaySuccess(self):
        # today
        moment_from = datetime.datetime.now()
        moment_to = moment_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 1)  # don't split

    def test_paginate_download_instrument_history_parameters_FiveYearsInsideIntervalSuccess(self):
        # 5 years ago inside interval
        moment_from = datetime.datetime.now() - datetime.timedelta(days=365*10)
        moment_to = datetime.datetime.now()
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 2)  # interval was split
        (first_params, _, _), (second_params, _, _) = paginated_parameters_tuples
        self.assertLessEqual(first_params.date_from, first_params.date_to)
        self.assertLessEqual(second_params.date_from, second_params.date_to)

        # even if parameters was disaligned
        long_ago = datetime.date(1945, 1, 1)
        parameters = self.generate_history_download_parameters(long_ago, long_ago)

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 2)  # interval was split
        (first_params, _, _), (second_params, _, _) = paginated_parameters_tuples
        self.assertLessEqual(first_params.date_from, first_params.date_to)
        self.assertLessEqual(second_params.date_from, second_params.date_to)

        today = datetime.date.today()
        parameters = self.generate_history_download_parameters(today, today)

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 2)  # interval was split
        (first_params, _, _), (second_params, _, _) = paginated_parameters_tuples
        self.assertLessEqual(first_params.date_from, first_params.date_to)
        self.assertLessEqual(second_params.date_from, second_params.date_to)

    def test_paginate_download_instrument_history_parameters_OneDayDontSplit(self):
        # one day don't split
        moment_from = moment_to = datetime.datetime.now()
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        paginated_parameters_tuples = \
            self.string_data_downloader.paginate_download_instrument_history_parameters(
                parameters,
                moment_from,
                moment_to)
        paginated_parameters_tuples = list(paginated_parameters_tuples)  # materialize generator

        self.assertEqual(len(paginated_parameters_tuples), 1)  # exactly one

    def test_download_index_history_string_Success(self):
        date_from = datetime.date(2010, 1, 1)
        date_to = date_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(date_from, date_to)
        index_id = parameters.index_id
        context = parameters.context
        index_level = parameters.index_level
        currency = parameters.currency

        result = self.string_data_downloader.download_index_history_string(
            index_id,
            context,
            index_level,
            currency,
            date_from,
            date_to)
        
        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instrument_history_string_Success(self):
        moment_from = datetime.datetime(2010, 1, 1)  # no hours
        moment_to = moment_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        result = self.string_data_downloader.download_instrument_history_string(parameters, moment_from, moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

        moment_from = datetime.datetime(2010, 1, 1, 12)  # has hours
        moment_to = moment_from + datetime.timedelta(days=365)
        parameters = self.generate_history_download_parameters(moment_from.date(), moment_to.date())

        result = self.string_data_downloader.download_instrument_history_string(parameters, moment_from, moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_indexes_info_string_Success(self):
        market = Markets.REGIONAL_ALL_COUNTRY
        context = Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL)
        
        result = self.string_data_downloader.download_indexes_info_string(market, context)
        
        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instruments_info_string_Success(self):
        market = Markets.REGIONAL_ALL_COUNTRY
        context = Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL)
        parameters = MsciIndexesInfoDownloadParameters.safe_create(market=market, context=context)

        result = self.string_data_downloader.download_instruments_info_string(parameters)

        self.assertEqual(result.downloaded_string, self.fake_data)


class TestMsciAPIActualityChecker(unittest.TestCase):
    
    def setUp(self):
        # fakes
        self.fake_info_str_data = 'info'
        self.fake_history_str_data = 'history'
        
        self.success_history_parsed_item = \
            IndexValue(
                date=MsciApiActualityChecker._expectedFirstDate,
                value=MsciApiActualityChecker._expectedFirstValue,
                index_name=MsciApiActualityChecker._expectedIndexName,
                style=MsciApiActualityChecker._expectedIndexContext.style,
                size=MsciApiActualityChecker._expectedIndexContext.size
            )
        self.success_info_parsed_item = IndexInfo(index_id='FAKE_ID', name='FAKE_NAME')
        
        self.fake_string_data_downloader = FakeMsciStringDataDownloader(
            self.fake_info_str_data,
            self.fake_history_str_data)
    
    def test_check_Success(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        checker.check()

        # check that there is no incorrect downloaded strings
        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.fake_string_data_downloader.download_instruments_info_string_results))
        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.fake_string_data_downloader.download_instrument_history_string_results))

    def test_check_raisesWhenInfoParseError(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        info_parser.parse_exception = ParseError('Error')

        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenInfoUnknownError(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        info_parser.parse_exception = Exception('Error')

        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenNoInfo(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([])  # No data
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryParseError(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        history_parser.parse_exception = ParseError('Error')
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])

        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryUnknownError(self):
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        history_parser.parse_exception = Exception('Error')
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])

        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongIndexName(self):
        # corrupt data
        self.success_history_parsed_item.index_name = 'WRONG'
        
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongIndexSize(self):
        # corrupt data
        self.success_history_parsed_item.size = None
        
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongIndexStyle(self):
        # corrupt data
        self.success_history_parsed_item.style = None
        
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongFirstDate(self):
        # corrupt data
        self.success_history_parsed_item.date = datetime.date.max
        
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()
    
    def test_check_raisesWhenWrongFirstValue(self):
        # corrupt data
        self.success_history_parsed_item.value = decimal.Decimal(-1)
        
        history_parser = FakeMsciHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeIndexInfoParser([self.success_info_parsed_item])
        
        checker = MsciApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)
        
        with self.assertRaises(CheckApiActualityError):
            checker.check()


class TestMsciIndexDownloadParameterValuesStorage(unittest.TestCase):

    def setUp(self):
        self.storage = MsciIndexDownloadParameterValuesStorage()

    def test_is_dynamic_enum_type_AlwaysFalse(self):
        self.assertFalse(self.storage.is_dynamic_enum_type(Markets))
        self.assertFalse(self.storage.is_dynamic_enum_type(Sizes))
        self.assertFalse(self.storage.is_dynamic_enum_type(IndexLevels))
        # noinspection PyTypeChecker
        self.assertFalse(self.storage.is_dynamic_enum_type(None))

    def test_get_dynamic_enum_key_AlwaysNone(self):
        self.assertIsNone(self.storage.get_dynamic_enum_key(Markets.REGIONAL_ALL_COUNTRY))
        self.assertFalse(self.storage.get_dynamic_enum_key(Sizes.REGIONAL_STANDARD))
        self.assertFalse(self.storage.get_dynamic_enum_key(IndexLevels.PRICE))

    def test_get_all_managed_types_AlwaysEmpty(self):
        self.assertSequenceEqual(list(self.storage.get_all_managed_types()), [])

    def test_get_parameter_type_choices_Success(self):
        all_types = self.storage._special_handlers.keys()
        for dynamic_enum_type in all_types:
            choices = self.storage.get_parameter_type_choices(dynamic_enum_type)

            self.assertIsNotNone(choices)

        # noinspection PyTypeChecker
        choices = self.storage.get_parameter_type_choices(None)

        self.assertIsNone(choices)

        choices = self.storage.get_parameter_type_choices(Styles)

        self.assertIsNone(choices)


class TestMsciIndexExporterFactory(CommonTestCases.CommonIndexExporterFactoryTests):

    def get_exporter_factory(self) -> InstrumentExporterFactory:
        return MsciIndexExporterFactory()

    def is_dynamic_enum_type_manager_singleton(self):
        return True

    def is_download_parameters_factory_singleton(self):
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
import datetime

from sane_finances.communication.downloader import DownloadStringResult
from sane_finances.sources.moex.v1_3.exporters import (
    MoexStringDataDownloader, MoexDownloadParameterValuesStorage)
from sane_finances.sources.moex.v1_3.meta import (
    SecurityValue, SecurityInfo, GlobalIndexData)
from sane_finances.sources.moex.v1_3.parsers import (
    MoexHistoryJsonParser, MoexGlobalIndexJsonParser, MoexSecurityInfoJsonParser)
from ....communication.fakes import FakeDownloader


class FakeMoexGlobalIndexJsonParser(MoexGlobalIndexJsonParser):

    def __init__(self, fake_global_index_data: GlobalIndexData):
        super().__init__()

        self.fake_global_index_data = fake_global_index_data
        self.parse_exception = None

    def parse(self, raw_json_text: str) -> GlobalIndexData:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_global_index_data


class FakeMoexDownloadParameterValuesStorage(MoexDownloadParameterValuesStorage):

    def __init__(self, fake_global_index_data: GlobalIndexData):
        super().__init__(FakeDownloader(None), FakeMoexGlobalIndexJsonParser(fake_global_index_data))


class FakeMoexSecurityInfoJsonParser(MoexSecurityInfoJsonParser):

    def __init__(self, fake_data: typing.Iterable[SecurityInfo]):
        # noinspection PyTypeChecker
        super().__init__(None)

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(self, xml) -> typing.Iterable[SecurityInfo]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeMoexHistoryJsonParser(MoexHistoryJsonParser):

    def __init__(self, fake_data: typing.Iterable[SecurityValue]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(
            self,
            raw_xml_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[SecurityValue]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeMoexStringDataDownloader(MoexStringDataDownloader):

    def __init__(self, fake_info_data, fake_history_data):
        super().__init__(FakeDownloader(None))

        self.download_instruments_info_string_results: typing.List[DownloadStringResult] = []
        self.download_instrument_history_string_results: typing.List[DownloadStringResult] = []

        self.fake_history_data = fake_history_data
        self.fake_info_data = fake_info_data

    def download_security_history_string(
            self,
            board,
            sec_id,
            start,
            date_from,
            date_to) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_history_data)
        self.download_instrument_history_string_results.append(result)
        return result

    def download_securities_info_string(
            self,
            board) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_info_data)
        self.download_instruments_info_string_results.append(result)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
import datetime

from sane_finances.communication.downloader import DownloadStringResult
from sane_finances.sources.cbr.v2016.meta import (
    CurrencyRateValue, CurrencyInfo)
from sane_finances.sources.cbr.v2016.parsers import (
    CbrCurrencyInfoParser, CbrCurrencyHistoryXmlParser)
from sane_finances.sources.cbr.v2016.exporters import (
    CbrStringDataDownloader)

from ....communication.fakes import FakeDownloader


class FakeCbrCurrencyInfoParser(CbrCurrencyInfoParser):

    def __init__(self, fake_data: typing.Iterable[CurrencyInfo]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(self, xml) -> typing.Iterable[CurrencyInfo]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeCbrCurrencyHistoryXmlParser(CbrCurrencyHistoryXmlParser):

    def __init__(self, fake_data: typing.Iterable[CurrencyRateValue]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(
            self,
            raw_xml_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[CurrencyRateValue]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeCbrStringDataDownloader(CbrStringDataDownloader):

    def __init__(self, fake_info_data, fake_history_data):
        super().__init__(FakeDownloader(None))

        self.download_instruments_info_string_results: typing.List[DownloadStringResult] = []
        self.download_instrument_history_string_results: typing.List[DownloadStringResult] = []

        self.fake_history_data = fake_history_data
        self.fake_info_data = fake_info_data

    def download_currency_history_string(
            self,
            currency_id,
            date_from,
            date_to) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_history_data)
        self.download_instrument_history_string_results.append(result)
        return result

    def download_currencies_info_string(self, rate_frequency) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_info_data)
        self.download_instruments_info_string_results.append(result)
        return result



""" All available sources for export instruments history data.
"""

from . import msci
from . import solactive
from . import moex
from . import cbr

# when new source appears, add it here in the same way

__all__ = ['msci', 'solactive', 'moex', 'cbr']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utilities for parse data from cbr.ru.
"""

import decimal
import logging
import typing
from xml.etree import ElementTree
import datetime

from .meta import CurrencyRateValue, CurrencyInfo
from ...base import InstrumentValuesHistoryParser, InstrumentInfoParser, ParseError

logging.getLogger().addHandler(logging.NullHandler())


def _get_xml_last_item_text(element: ElementTree.Element, path: str) -> typing.Optional[str]:
    # get last item
    text = None
    for value in element.iterfind(path):
        text = value.text
    return text


class CbrCurrencyHistoryXmlParser(InstrumentValuesHistoryParser):
    """ Parser for history data of currency from XML string.

    E.g.::

        <ValCurs ID="R01235" DateRange1="02.03.2001" DateRange2="14.03.2001" name="Foreign Currency Market Dynamic">
            <Record Date="02.03.2001" Id="R01235">
                <Nominal>1</Nominal>
                <Value>28,6200</Value>
            </Record>
            ...
        </ValCurs>
    """
    RootTag = 'ValCurs'
    date_format = '%d.%m.%Y'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(  # pylint: disable=arguments-renamed
            self,
            raw_xml_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[CurrencyRateValue]:

        try:
            root = ElementTree.fromstring(raw_xml_text)
        except ElementTree.ParseError as ex:
            raise ParseError(ex.msg) from ex

        expected_tag = self.RootTag
        if root.tag != expected_tag:
            raise ParseError(f"Wrong XML format. Root ('{root.tag}') is not '{expected_tag}'.")

        id_tag = 'Id'
        date_tag = 'Date'

        for record_element in root.iterfind('./Record'):
            if id_tag not in record_element.attrib:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found {id_tag} attrib in\n{ElementTree.tostring(record_element)}")
            if date_tag not in record_element.attrib:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found {date_tag} attrib in\n{ElementTree.tostring(record_element)}")

            currency_id: str = record_element.attrib[id_tag]
            self.logger.debug(f"Got currency '{currency_id}'")

            date_raw_text: str = record_element.attrib[date_tag]
            value_raw_text = _get_xml_last_item_text(record_element, './Value')
            nominal_raw_text = _get_xml_last_item_text(record_element, './Nominal')

            self.logger.debug(f"Got {date_raw_text!r} -> {value_raw_text!r}/{nominal_raw_text!r}")

            if nominal_raw_text is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found Nominal tag in\n{ElementTree.tostring(record_element)}")
            if value_raw_text is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found Value tag in\n{ElementTree.tostring(record_element)}")

            value_date = datetime.datetime.strptime(date_raw_text, self.date_format).date()
            value = decimal.Decimal(value_raw_text.replace(',', '.'))
            nominal = int(nominal_raw_text)

            yield CurrencyRateValue(date=value_date, value=value, nominal=nominal, currency_id=currency_id)


class CbrCurrencyInfoParser(InstrumentInfoParser):
    """ Parser for indexes info list from XML.

    E.g.::

        <Valuta name="Foreign Currency Market Lib">
            <Item ID="R01010">
                <Name>Австралийский доллар</Name>
                <EngName>Australian Dollar</EngName>
                <Nominal>1</Nominal>
                <ParentCode>R01010 </ParentCode>
            </Item>
            ...
        </Valuta>
    """

    RootTag = 'Valuta'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(self, raw_xml_text: str) -> typing.Iterable[CurrencyInfo]:  # pylint: disable=arguments-renamed
        try:
            root = ElementTree.fromstring(raw_xml_text)
        except ElementTree.ParseError as ex:
            raise ParseError(ex.msg) from ex

        expected_tag = self.RootTag
        if root.tag != expected_tag:
            raise ParseError(f"Wrong XML format. Root ('{root.tag}') is not '{expected_tag}'.")

        id_tag = 'ID'

        has_any = False
        for item_element in root.iterfind('./Item'):

            if id_tag not in item_element.attrib:
                self.logger.error(f"Currency ID not found in\n{ElementTree.tostring(item_element)}")
                raise ParseError("Unexpected XML format. Currency ID not found")

            currency_id = item_element.attrib[id_tag]
            name = _get_xml_last_item_text(item_element, './Name')
            eng_name = _get_xml_last_item_text(item_element, './EngName')
            nominal_raw_text = _get_xml_last_item_text(item_element, './Nominal')
            parent_code = _get_xml_last_item_text(item_element, './ParentCode')

            if name is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found Name tag in\n{ElementTree.tostring(item_element)}")
            if eng_name is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found EngName tag in\n{ElementTree.tostring(item_element)}")
            if nominal_raw_text is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found Nominal tag in\n{ElementTree.tostring(item_element)}")
            if parent_code is None:
                raise ParseError(f"Wrong XML format. "
                                 f"Not found ParentCode tag in\n{ElementTree.tostring(item_element)}")

            parent_code = parent_code.strip()
            nominal = int(nominal_raw_text)

            has_any = True
            yield CurrencyInfo(
                currency_id=currency_id,
                name=name,
                eng_name=eng_name,
                nominal=nominal,
                parent_code=parent_code)

        if not has_any:
            self.logger.error(f"No data found in XML:\n{raw_xml_text}")
            raise ParseError("Unexpected XML format. No data found")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import InstrumentValue, InstrumentInfo, DownloadParametersFactory, \
    InstrumentHistoryDownloadParameters
from sane_finances.sources.msci.v1.meta import (
    Scopes, Formats, Context,
    IndexValue, IndexInfo, MsciIndexHistoryDownloadParameters, MsciIndexesInfoDownloadParameters,
    MsciDownloadParametersFactory, Styles, Sizes, Markets, IndexLevels, Currencies)
from .common import CommonTestCases


class TestFormats(unittest.TestCase):

    def test_get_file_extension_Success(self):
        for value in Formats:
            _ = Formats.get_file_extension(value)


class TestIndexValue(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_value = IndexValue(
            date=datetime.date(2000, 12, 31),
            value=decimal.Decimal(42),
            index_name='NAME',
            style=Styles.NONE,
            size=Sizes.REGIONAL_STANDARD)
        expected_instrument_value = InstrumentValue(
            value=index_value.value,
            moment=datetime.datetime.combine(index_value.date, datetime.time.min, tzinfo=datetime.timezone.utc))

        instrument_value = index_value.get_instrument_value(datetime.timezone.utc)

        self.assertEqual(expected_instrument_value, instrument_value)

    def test_raiseWrongDate(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexValue(
                date=None,
                value=decimal.Decimal(42),
                index_name='NAME',
                style=Styles.NONE,
                size=Sizes.REGIONAL_STANDARD)


class TestIndexInfo(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_info = IndexInfo(
            index_id='CODE',
            name='NAME')
        expected_instrument_info = InstrumentInfo(code=index_info.index_id, name=index_info.name)

        instrument_info = index_info.instrument_info

        self.assertEqual(expected_instrument_info, instrument_info)


class TestMsciIndexesInfoDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = MsciIndexesInfoDownloadParameters.safe_create(
            market=Markets.COUNTRY_DEVELOPED_MARKETS,
            context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL))

    def test_safe_create_raiseWrongContext(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexesInfoDownloadParameters.safe_create(
                market=Markets.COUNTRY_DEVELOPED_MARKETS,
                context=None)


class TestMsciIndexHistoryDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = MsciIndexHistoryDownloadParameters.safe_create(
            index_id='CODE',
            context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL),
            index_level=IndexLevels.PRICE,
            currency=Currencies.USD,
            date_from=datetime.date(2000, 12, 31),
            date_to=datetime.date(2000, 12, 31))

    def test_safe_create_raiseWrongContext(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_id='CODE',
                context=None,
                index_level=IndexLevels.PRICE,
                currency=Currencies.USD,
                date_from=datetime.date(2000, 12, 31),
                date_to=datetime.date(2000, 12, 31))

    def test_safe_create_raiseWrongDateFrom(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_id='CODE',
                context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL),
                index_level=IndexLevels.PRICE,
                currency=Currencies.USD,
                date_from=None,
                date_to=datetime.date(2000, 12, 31))

    def test_safe_create_raiseWrongDateTo(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_id='CODE',
                context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL),
                index_level=IndexLevels.PRICE,
                currency=Currencies.USD,
                date_from=datetime.date(2000, 12, 31),
                date_to=None)

    def test_safe_create_raiseWrongDates(self):
        date_from = datetime.date(2000, 12, 31)
        date_to = date_from - datetime.timedelta(days=1)

        with self.assertRaises(ValueError):
            _ = MsciIndexHistoryDownloadParameters.safe_create(
                index_id='CODE',
                context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL),
                index_level=IndexLevels.PRICE,
                currency=Currencies.USD,
                date_from=date_from,
                date_to=date_to)

    def test_clone_with_instrument_info_parameters_Success(self):
        params = MsciIndexHistoryDownloadParameters(
            index_id='CODE',
            context=Context(style=Styles.NONE, size=Sizes.REGIONAL_STANDARD, scope=Scopes.REGIONAL),
            index_level=IndexLevels.PRICE,
            currency=Currencies.USD,
            date_from=datetime.date(2000, 12, 31),
            date_to=datetime.date(2000, 12, 31))
        expected_result = params

        result = params.clone_with_instrument_info_parameters(None, None)

        self.assertEqual(expected_result, result)


class TestMsciDownloadParametersFactory(CommonTestCases.CommonDownloadParametersFactoryTests):

    def get_download_parameters_factory(self) -> DownloadParametersFactory:
        return MsciDownloadParametersFactory()

    def get_generate_history_download_parameters_from_expected_result_with_none(self) \
            -> InstrumentHistoryDownloadParameters:
        # noinspection PyTypeChecker
        expected_result = MsciIndexHistoryDownloadParameters(
            index_id=None,
            context=None,
            index_level=None,
            currency=None,
            date_from=None,
            date_to=None)
        return expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import datetime
import decimal

from sane_finances.sources.base import CheckApiActualityError, ParseError, InstrumentExporterFactory, SourceError
from sane_finances.sources.moex.v1_3.exporters import (
    MoexDynamicEnumTypeManager, MoexDownloadParameterValuesStorage,
    MoexStringDataDownloader, MoexApiActualityChecker, MoexIndexExporterFactory_v1_3)
from sane_finances.sources.moex.v1_3.meta import (
    MoexSecuritiesInfoDownloadParameters,
    MoexSecurityHistoryDownloadParameters, TradeEngine, Market, Board, GlobalIndexData, SecurityInfo, SecurityValue)

from .fakes import (
    FakeDownloader, FakeMoexStringDataDownloader, FakeMoexDownloadParameterValuesStorage, FakeMoexHistoryJsonParser,
    FakeMoexGlobalIndexJsonParser, FakeMoexSecurityInfoJsonParser)
from .common import CommonTestCases


class TestMoexDynamicEnumTypeManager(unittest.TestCase):

    def test_is_dynamic_enum_type_Success(self):
        manager = MoexDynamicEnumTypeManager()
        all_types = manager.get_all_managed_types()

        for dynamic_enum_type in all_types:
            self.assertTrue(manager.is_dynamic_enum_type(dynamic_enum_type))

    def test_is_dynamic_enum_type_FalseWithWrongType(self):
        manager = MoexDynamicEnumTypeManager()

        # noinspection PyTypeChecker
        self.assertFalse(manager.is_dynamic_enum_type(None))

    def test_get_dynamic_enum_key_Success(self):
        manager = MoexDynamicEnumTypeManager()
        engine = TradeEngine(identity=42, name='NAME', title='TITLE')

        self.assertEqual(manager.get_dynamic_enum_key(engine), engine.identity)

    def test_get_dynamic_enum_key_NoneWithWrongType(self):
        manager = MoexDynamicEnumTypeManager()

        self.assertIsNone(manager.get_dynamic_enum_key(None))


class TestMoexDownloadParameterValuesStorage(unittest.TestCase):

    def setUp(self):
        engine = TradeEngine(
            identity=42,
            name='stock',
            title='Фондовый рынок и рынок депозитов')
        market = Market(
            identity=42,
            trade_engine=engine,
            name='shares',
            title='Рынок акций',
            marketplace='MXSE')
        board = Board(
            identity=42,
            trade_engine=engine,
            market=market,
            boardid='TQTF',
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        self.global_index_data = GlobalIndexData(
            trade_engines=(engine, engine),
            markets=(market, market),
            boards=(board, board))

        self.downloader = FakeDownloader(None)
        self.global_index_json_parser = FakeMoexGlobalIndexJsonParser(self.global_index_data)
        self.storage = MoexDownloadParameterValuesStorage(
            self.downloader,
            self.global_index_json_parser)

    def test_reload_Success(self):
        self.storage.reload()

        self.assertGreaterEqual(len(self.downloader.download_string_results), 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.downloader.download_string_results))

    def test_reload_LastDownloadStringResultIsNotCorrectWhenParseError(self):
        self.global_index_json_parser.parse_exception = ParseError('Error')

        with self.assertRaises(Exception):
            self.storage.reload()

        self.assertGreaterEqual(len(self.downloader.download_string_results), 1)
        self.assertIs(self.downloader.download_string_results[-1].is_correct, False)

    def test_reload_LastDownloadStringResultIsNotCorrectWhenUnknownError(self):
        self.global_index_json_parser.parse_exception = Exception('Error')

        with self.assertRaises(Exception):
            self.storage.reload()

        self.assertGreaterEqual(len(self.downloader.download_string_results), 1)
        self.assertIs(self.downloader.download_string_results[-1].is_correct, False)

    def test_get_dynamic_enum_value_by_key_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_dynamic_enum_value_by_key(dynamic_enum_type, 42)
            wrong_value = self.storage.get_dynamic_enum_value_by_key(dynamic_enum_type, -1)

            self.assertIsNotNone(value)
            self.assertIsNone(wrong_value)

        # noinspection PyTypeChecker
        value = self.storage.get_dynamic_enum_value_by_key(None, 42)

        self.assertIsNone(value)

    def test_get_dynamic_enum_value_by_choice_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_dynamic_enum_value_by_choice(dynamic_enum_type, '42')
            wrong_value = self.storage.get_dynamic_enum_value_by_choice(dynamic_enum_type, '-1')

            self.assertIsNotNone(value)
            self.assertIsNone(wrong_value)

        # noinspection PyTypeChecker
        value = self.storage.get_dynamic_enum_value_by_choice(None, '42')

        self.assertIsNone(value)

    def test_get_dynamic_enum_value_by_choice_RaiseWhenWrongChoice(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            with self.assertRaises(SourceError):
                _ = self.storage.get_dynamic_enum_value_by_choice(dynamic_enum_type, 'WRONG')

    def test_get_all_parameter_values_for_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_all_parameter_values_for(dynamic_enum_type)

            self.assertIsNotNone(value)

        # noinspection PyTypeChecker
        value = self.storage.get_all_parameter_values_for(None)

        self.assertIsNone(value)

    def test_get_parameter_type_choices_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            choices = self.storage.get_parameter_type_choices(dynamic_enum_type)

            self.assertIsNotNone(choices)

        # noinspection PyTypeChecker
        choices = self.storage.get_parameter_type_choices(None)

        self.assertIsNone(choices)


class TestMoexStringDataDownloader(unittest.TestCase):

    def setUp(self):
        self.fake_data = 'data'

        self.engine = TradeEngine(
            identity=42,
            name='stock',
            title='Фондовый рынок и рынок депозитов')
        self.market = Market(
            identity=42,
            trade_engine=self.engine,
            name='shares',
            title='Рынок акций',
            marketplace='MXSE')
        self.board = Board(
            identity=42,
            trade_engine=self.engine,
            market=self.market,
            boardid='TQTF',
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)

        self.string_data_downloader = MoexStringDataDownloader(FakeDownloader(self.fake_data))
        self.history_download_params = MoexSecurityHistoryDownloadParameters(board=self.board, sec_id='SECID', start=0)

    def test_paginate_download_instrument_history_parameters_Success(self):
        params = MoexSecurityHistoryDownloadParameters(self.board, sec_id='SEC ID', start=0)
        moment_from = datetime.datetime(2010, 1, 1)
        moment_to = moment_from + datetime.timedelta(days=1)
        test_steps = 10  # number of first pages to test

        self.assertGreaterEqual(test_steps, 2)

        current_step = 1
        current_start = 0  # start from zero
        for paginated_params, paginated_moment_from, paginated_moment_to in \
                self.string_data_downloader.paginate_download_instrument_history_parameters(
                    params, moment_from, moment_to):
            self.assertEqual(paginated_moment_from, moment_from)
            self.assertEqual(paginated_moment_to, moment_to)
            self.assertEqual(paginated_params.board, params.board)
            self.assertEqual(paginated_params.sec_id, params.sec_id)
            self.assertEqual(paginated_params.start, current_start)

            current_start += self.string_data_downloader.limit.value
            current_step += 1
            if current_step > test_steps:
                # limit pages for test purpose because pagination in this exporter is infinite by design
                break

    def test_download_instrument_history_string_Success(self):
        moment_from = datetime.datetime(2010, 1, 1, 12)  # has hours
        moment_to = moment_from + datetime.timedelta(days=1)

        result = self.string_data_downloader.download_instrument_history_string(
            self.history_download_params,
            moment_from,
            moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

        moment_from = datetime.datetime(2010, 1, 1)  # no hours
        moment_to = moment_from + datetime.timedelta(days=1)

        result = self.string_data_downloader.download_instrument_history_string(
            self.history_download_params,
            moment_from,
            moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_security_history_string_Success(self):
        date_from = datetime.date(2010, 1, 1)
        date_to = date_from + datetime.timedelta(days=1)

        result = self.string_data_downloader.download_security_history_string(
            self.history_download_params.board,
            self.history_download_params.sec_id,
            self.history_download_params.start,
            date_from,
            date_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instruments_info_string_Success(self):
        params = MoexSecuritiesInfoDownloadParameters.safe_create(board=self.board)

        result = self.string_data_downloader.download_instruments_info_string(params)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_securities_info_string_Success(self):
        result = self.string_data_downloader.download_securities_info_string(self.board)

        self.assertEqual(result.downloaded_string, self.fake_data)


class TestMoexApiActualityChecker(unittest.TestCase):

    def setUp(self):
        self.engine = TradeEngine(
            identity=42,
            name=MoexApiActualityChecker._trade_engine_name_to_test,
            title='Фондовый рынок и рынок депозитов')
        self.market = Market(
            identity=42,
            trade_engine=self.engine,
            name=MoexApiActualityChecker._market_name_to_test,
            title='Рынок акций',
            marketplace='MXSE')
        self.board = Board(
            identity=42,
            trade_engine=self.engine,
            market=self.market,
            boardid=MoexApiActualityChecker._boardid_to_test,
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        self.global_index_data = GlobalIndexData(
            trade_engines=(self.engine,),
            markets=(self.market,),
            boards=(self.board,))

        self.success_info_data = [SecurityInfo(
            sec_id='SEC ID',
            board=self.board,
            short_name='SHORT NAME')
        ]
        self.index_info_parser = FakeMoexSecurityInfoJsonParser(self.success_info_data)

        self.success_history_data = [SecurityValue(
            trade_date=MoexApiActualityChecker._history_date_to_test,
            close=MoexApiActualityChecker._expected_close_value)
        ]
        self.history_parser = FakeMoexHistoryJsonParser(self.success_history_data)

        self.string_data_downloader = FakeMoexStringDataDownloader(None, None)

    def get_checker(self):
        return MoexApiActualityChecker(
            FakeMoexDownloadParameterValuesStorage(self.global_index_data),
            self.string_data_downloader,
            self.index_info_parser,
            self.history_parser)

    def test_check_Success(self):
        checker = self.get_checker()
        checker.check()

        # check that there is no incorrect downloaded strings
        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instruments_info_string_results))
        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instrument_history_string_results))

    def test_check_SuccessWithUnknownBoard(self):
        unknown_board = Board(
            identity=42,
            trade_engine=TradeEngine(identity=42, name='UNKNOWN NAME', title='UNKNOWN TITLE'),
            market=self.market,
            boardid=MoexApiActualityChecker._boardid_to_test,
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        self.global_index_data = self.global_index_data._replace(boards=(unknown_board,))
        checker = self.get_checker()

        checker.check()

        # check that there is no incorrect downloaded strings
        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instruments_info_string_results))
        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instrument_history_string_results))

    def test_check_raisesWhenNoBoards(self):
        # corrupt data
        self.global_index_data = self.global_index_data._replace(boards=())
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenNotFoundTargetBoard(self):
        # corrupt data
        unknown_board = Board(
            identity=42,
            trade_engine=self.engine,
            market=self.market,
            boardid='UNKNOWN BOARDID',
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        self.global_index_data = self.global_index_data._replace(boards=(unknown_board,))
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenInfoParseError(self):
        # corrupt data
        self.index_info_parser.parse_exception = ParseError('Error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenInfoUnknownError(self):
        # corrupt data
        self.index_info_parser.parse_exception = Exception('Error')
        checker = self.get_checker()

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryParseError(self):
        # corrupt data
        self.history_parser.parse_exception = ParseError('Error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryUnknownError(self):
        # corrupt data
        self.history_parser.parse_exception = Exception('Error')
        checker = self.get_checker()

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenNoHistory(self):
        # corrupt data
        self.history_parser.fake_data = []  # No data
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongFirstDate(self):
        # corrupt data
        self.success_history_data[0].trade_date = datetime.date.max
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongFirstValue(self):
        # corrupt data
        self.success_history_data[0].close = decimal.Decimal(-1)
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)


# noinspection PyPep8Naming
class TestMoexIndexExporterFactory_v1_3(CommonTestCases.CommonIndexExporterFactoryTests):

    def get_exporter_factory(self) -> InstrumentExporterFactory:
        return MoexIndexExporterFactory_v1_3()

    def is_dynamic_enum_type_manager_singleton(self):
        return True

    def is_download_parameters_factory_singleton(self):
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import unittest.mock
import datetime
import urllib.error

from sane_finances.communication.url_downloader import UrlDownloader, UrlDownloadStringResult
from sane_finances.communication.downloader import DownloadError

from .fakes import FakeDummyCacher, FakeExpirableCacher


class TestUrlDownloadStringResult(unittest.TestCase):

    def setUp(self):
        self.cacher = FakeDummyCacher()
        self.result = UrlDownloadStringResult('OK', self.cacher, '', [], {})

    def test_set_correctness_DontDropWhenTrue(self):
        self.result.set_correctness(True)

        self.assertEqual(self.cacher.drop_count, 0)

    def test_set_correctness_DropWhenFalse(self):
        self.result.set_correctness(False)

        self.assertEqual(self.cacher.drop_count, 1)


class TestUrlDownloader(unittest.TestCase):

    def setUp(self):
        self.http_response_read_count = 0

        self.cacher = FakeExpirableCacher(initial_expiry=datetime.timedelta(days=1))
        self.downloader = UrlDownloader(self.cacher)

        # fake reader
        self.fake_data = 'data'
        self.http_response_error = None
        self.encoding = self.downloader.default_encoding
        mock_urlopen_patcher = unittest.mock.patch(
            'urllib.request.urlopen',
            **{'return_value.__enter__.return_value.read': self.fake_http_response_read})
        mock_urlopen_patcher.start()
        self.addCleanup(mock_urlopen_patcher.stop)

    def fake_http_response_read(self):
        self.http_response_read_count += 1

        if self.http_response_error is not None:
            raise self.http_response_error

        return self.fake_data.encode(self.encoding)

    def test_download_string_SuccessWithParameters(self):
        url = "http://localhost"
        self.downloader.parameters = [("some_param", "param_value")]
        self.downloader.headers = {"some_header": "header_value"}

        result = self.downloader.download_string(url)

        self.assertEqual(result.downloaded_string, self.fake_data)
        self.assertEqual(self.http_response_read_count, 1)

    def test_download_string_SuccessWithoutParameters(self):
        url = "http://localhost"

        result = self.downloader.download_string(url)

        self.assertEqual(result.downloaded_string, self.fake_data)
        self.assertEqual(self.http_response_read_count, 1)

    def test_download_string_UnicodeStringSuccess(self):
        url = "http://localhost"
        self.fake_data = 'Привет'  # unicode string with russian 'Hello'
        self.encoding = 'cp1251'

        result = self.downloader.download_string(url, encoding=self.encoding)

        self.assertEqual(result.downloaded_string, self.fake_data)
        self.assertEqual(self.http_response_read_count, 1)

    def test_download_string_CacheSuccess(self):
        url = "http://localhost"

        # first download
        result = self.downloader.download_string(url)

        self.assertEqual(result.downloaded_string, self.fake_data)
        self.assertEqual(self.http_response_read_count, 1)

        # verify that cacher expiry was initialised properly and not equal to default zero-delta
        self.assertNotEqual(self.cacher.expiry, self.cacher.default_expiry)

        # second download should result the same but without web trip (i.e. should be from cache)
        result = self.downloader.download_string(url)

        self.assertEqual(result.downloaded_string, self.fake_data)
        self.assertEqual(self.http_response_read_count, 1)  # didn't changed

    def test_download_string_RaiseProperExceptionOnErrors(self):
        url = "http://localhost"
        self.http_response_error = urllib.error.URLError("Some error")

        with self.assertRaises(DownloadError):
            _ = self.downloader.download_string(url)
